#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.vm.tx.unsigned: unsigned transaction methods for VM chains
"""

from ....util import msg, msg_r, die

class Unsigned:
	desc = 'unsigned transaction'

	# Return signed object or False. Don’t exit or raise exception:
	async def sign(self, keys, tx_num_str=''):

		from ....exception import TransactionChainMismatch
		try:
			self.check_correct_chain()
		except TransactionChainMismatch:
			return False

		o = self.txobj

		def do_mismatch_err(io, j, k, desc):
			m = 'A compromised online installation may have altered your serialized data!'
			fs = '\n{} mismatch!\n{}\n  orig:       {}\n  serialized: {}'
			die(3, fs.format(desc.upper(), m, getattr(io[0], k), o[j]))

		if o['from'] != self.inputs[0].addr:
			do_mismatch_err(self.inputs, 'from', 'addr', 'from-address')
		if self.outputs:
			if o['to'] != self.outputs[0].addr:
				do_mismatch_err(self.outputs, 'to', 'addr', 'to-address')
			if o['amt'] != self.outputs[0].amt:
				do_mismatch_err(self.outputs, 'amt', 'amt', 'amount')

		msg_r(f'Signing transaction{tx_num_str}...')

		try:
			await self.do_sign(o, keys[0].sec.wif)
			msg('OK')
			from ....tx import SignedTX
			tx = SignedTX(cfg=self.cfg, data=self.__dict__, automount=self.automount)
			tx.check_serialized_integrity()
			return tx
		except Exception as e:
			msg(f'{e}: transaction signing failed!')
			return False



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
tool.help: Help screen routines for the 'mmgen-tool' utility
"""

from .. import main_tool

from .common import tool_cmd_base

def main_help():

	from ..util import capfirst
	from ..util2 import pretty_format

	def do():
		for clsname, cmdlist in main_tool.mods.items():
			cls = main_tool.get_mod_cls(clsname)
			cls_docstr = cls.__doc__.strip()
			yield capfirst(cls_docstr.split('\n')[0].strip()) + ':'
			yield ''

			if '\n' in cls_docstr:
				for line in cls_docstr.split('\n')[2:]:
					yield '  ' + line.lstrip('\t')
				yield ''

			max_w = max(map(len, cmdlist))

			for cmdname in cmdlist:
				code = getattr(cls, cmdname)
				if code.__doc__:
					yield '  {:{}} - {}'.format(
						cmdname,
						max_w,
						pretty_format(
							code.__doc__.strip().split('\n')[0].strip(),
							width = 79-(max_w+7),
							pfx   = ' '*(max_w+5)).lstrip()
					)
			yield ''

	return '\n'.join(do())

def gen_tool_usage():

	from ..util import capfirst

	m1 = """
		GENERAL USAGE INFORMATION FOR MMGEN-TOOL COMMANDS

		  Arguments with only type specified in square brackets are required

		  Arguments with both type and default value specified in square brackets are
		  optional and must be specified in the form ‘name=value’

		  For more detailed usage information for a particular tool command, type
		  ‘mmgen-tool help <command name>’
		"""

	m2 = """
		  To force a command to read from STDIN instead of file (for commands taking
		  a filename as their first argument), substitute "-" for the filename.


		EXAMPLES:

		  Generate a random LTC Bech32 public/private keypair:
		  $ mmgen-tool -r0 --coin=ltc --type=bech32 randpair

		  Generate a DASH address with compressed public key from the supplied WIF key:
		  $ mmgen-tool --coin=dash --type=compressed wif2addr XJkVRC3eGKurc9Uzx1wfQoio3yqkmaXVqLMTa6y7s3M3jTBnmxfw

		  Generate a well-known burn address:
		  $ mmgen-tool hextob58chk 000000000000000000000000000000000000000000

		  Generate a random 12-word seed phrase:
		  $ mmgen-tool -r0 mn_rand128 fmt=bip39

		  Same as above, but get additional entropy from user:
		  $ mmgen-tool mn_rand128 fmt=bip39

		  Encode bytes from a file to base 58:
		  $ mmgen-tool bytestob58 /etc/timezone pad=20

		  Reverse a hex string:
		  $ mmgen-tool hexreverse "deadbeefcafe"

		  Same as above, but supply input via STDIN:
		  $ echo "deadbeefcafe" | mmgen-tool hexreverse -
		"""

	for line in m1.lstrip().split('\n'):
		yield line.lstrip('\t')

	for clsname, cmdlist in main_tool.mods.items():
		cls = main_tool.get_mod_cls(clsname)
		cls_docstr = cls.__doc__.strip()
		yield ''
		yield '  {}:'.format(capfirst(cls_docstr.split('\n')[0].strip()))
		yield ''

		if '\n' in cls_docstr:
			for line in cls_docstr.split('\n')[2:]:
				yield '    ' + line.lstrip('\t')
			yield ''

		max_w = max(map(len, cmdlist))
		for cmdname in cmdlist:
			yield '    {a:{w}} {b}'.format(
				a = cmdname,
				b = main_tool.create_call_sig(cmdname, cls, as_string=True),
				w = max_w)
		yield ''

	for line in m2.rstrip().split('\n'):
		yield line.lstrip('\t')

def gen_tool_cmd_usage(mod, cmdname):

	from ..cfg import gc
	from ..util import capfirst

	cls = main_tool.get_mod_cls(mod)
	docstr = getattr(cls, cmdname).__doc__.strip()
	args, kwargs, kwargs_types, _, ann = main_tool.create_call_sig(cmdname, cls)
	ARGS = 'ARG' if len(args) == 1 else 'ARGS' if args else ''
	KWARGS = 'KEYWORD ARG' if len(kwargs) == 1 else 'KEYWORD ARGS' if kwargs else ''

	yield capfirst(docstr.split('\n')[0].strip())
	yield ''
	yield 'USAGE: {b} [OPTS] {c}{d}{e}'.format(
		b = gc.prog_name,
		c = cmdname,
		d = f' {ARGS}' if ARGS else '',
		e = f' [{KWARGS}]' if KWARGS else '')

	if args:
		max_w = max(len(k[0]) for k in args)
		yield ''
		yield f'Required {ARGS} (type shown in square brackets):'
		yield ''
		for argname, argtype in args:
			have_sstr = ann.get(argname) == 'sstr'
			yield '  {a:{w}} [{b}]{c}{d}'.format(
				a = argname,
				b = argtype,
				c = " (use '-' to read from STDIN)" if have_sstr else '',
				d = ' ' + ann[argname] if isinstance(ann.get(argname), str) and not have_sstr else '',
				w = max_w)

	if kwargs:
		max_w = max(len(k) for k in kwargs)
		max_w2 = max(len(kwargs_types[k].__name__) + len(repr(kwargs[k])) for k in kwargs) + 3
		yield ''
		yield f'Optional {KWARGS} (type and default value shown in square brackets):'
		yield ''
		for argname in kwargs:
			yield '  {a:{w}} {b:{w2}} {c}'.format(
				a = argname,
				b = '[{}={}]'.format(kwargs_types[argname].__name__, repr(kwargs[argname])),
				c = capfirst(ann[argname]) if isinstance(ann.get(argname), str) else '',
				w = max_w,
				w2 = max_w2).rstrip()

	if '\n' in docstr:
		for line in docstr.split('\n')[1:]:
			yield line.lstrip('\t')

def usage(cmdname=None, *, exit_val=1):

	from ..util import Msg, die

	if cmdname:
		for mod, cmdlist in main_tool.mods.items():
			if cmdname in cmdlist:
				Msg('\n'.join(gen_tool_cmd_usage(mod, cmdname)))
				break
		else:
			die(1, f'{cmdname!r}: no such tool command')
	else:
		from ..ui import do_pager
		do_pager('\n'.join(gen_tool_usage()) + '\n')

	import sys
	sys.exit(exit_val)

class tool_cmd(tool_cmd_base):
	"help/usage commands"

	def help(self, command_name=''):
		"display usage information for a single command or all commands"
		usage(command_name, exit_val=0)

	def usage(self, command_name=''):
		"display usage information for a single command or all commands"
		usage(command_name, exit_val=0)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.xmr.daemon: Monero base protocol daemon classes
"""

import sys, os

from ...cfg import gc
from ...util import list_gen, die, contains_any
from ...daemon import CoinDaemon, RPCDaemon, _nw, _dd

class monero_daemon(CoinDaemon):
	daemon_data = _dd('Monero', 18004002, '0.18.4.2-release')
	networks = ('mainnet', 'testnet')
	exec_fn = 'monerod'
	testnet_dir = 'stagenet'
	new_console_mswin = True
	rpc_ports = _nw(18081, 38081, None) # testnet is stagenet
	cfg_file = 'bitmonero.conf'
	datadirs = {
		'linux': [gc.home_dir, '.bitmonero'],
		'darwin': [gc.home_dir, '.bitmonero'],
		'win32': ['/', 'c', 'ProgramData', 'bitmonero']
	}

	def init_datadir(self):
		self.logdir = super().init_datadir()
		return os.path.join(
			self.logdir,
			self.testnet_dir if self.network == 'testnet' else '')

	def get_p2p_port(self):
		return self.rpc_port - 1

	def init_subclass(self):

		from .rpc import MoneroRPCClient
		self.rpc = MoneroRPCClient(
			cfg    = self.cfg,
			proto  = self.proto,
			host   = 'localhost',
			port   = self.rpc_port,
			user   = None,
			passwd = None,
			test_connection = False,
			daemon = self)

		self.use_pidfile = sys.platform == 'linux'

		self.shared_args = list_gen(
			['--no-zmq'],
			[f'--p2p-bind-port={self.p2p_port}', self.p2p_port],
			[f'--rpc-bind-port={self.rpc_port}'],
			['--stagenet', self.network == 'testnet'],
		)

		self.coind_args = list_gen(
			['--hide-my-port'],
			['--no-igd'],
			[f'--data-dir={self.datadir}', self.non_dfl_datadir],
			[f'--pidfile={self.pidfile}', self.use_pidfile],
			['--detach',                  not (self.opt.no_daemonize or self.platform=='win32')],
			['--offline',                 not self.opt.online],
		)

	@property
	def stop_cmd(self):
		if self.platform == 'win32':
			return ['kill', '-Wf', self.pid]
		elif contains_any(self.start_cmd, ['--restricted-rpc', '--public-node']):
			return ['kill', self.pid]
		else:
			return [self.exec_fn] + self.shared_args + ['exit']

class MoneroWalletDaemon(RPCDaemon):

	master_daemon = 'monero_daemon'
	rpc_desc = 'Monero wallet'
	exec_fn = 'monero-wallet-rpc'
	coin = 'XMR'
	new_console_mswin = True
	networks = ('mainnet', 'testnet')
	rpc_ports = _nw(13131, 13141, None) # testnet is non-standard
	_reset_ok = ('debug', 'wait', 'pids', 'force_kill')
	test_suite_datadir = os.path.join('test', 'daemons', 'xmrtest', 'wallet_rpc')

	def start(self, *args, **kwargs):
		try: # NB: required due to bug in v18.3.1: PID file not deleted on shutdown
			os.unlink(self.pidfile)
		except FileNotFoundError:
			pass
		super().start(*args, **kwargs)

	def __init__(
			self,
			cfg,
			proto,
			*,
			wallet_dir  = None,
			test_suite  = False,
			user        = None,
			passwd      = None,
			monerod_addr = None,
			proxy       = None,
			port_shift  = None,
			datadir     = None,
			trust_monerod = False,
			test_monerod = False,
			opts         = None,
			flags        = None):

		self.proto = proto
		self.test_suite = test_suite

		super().__init__(cfg, opts=opts, flags=flags)

		self.network = proto.network
		self.wallet_dir = wallet_dir or (self.test_suite_datadir if test_suite else None)
		self.rpc_port = (
			self.cfg.wallet_rpc_port or
			getattr(self.rpc_ports, self.network) + (11 if test_suite else 0))
		if port_shift:
			self.rpc_port += port_shift

		id_str = f'{self.exec_fn}-{self.bind_port}'
		self.datadir = datadir or (self.test_suite_datadir if test_suite else self.exec_fn + '.d')
		self.pidfile = os.path.join(self.datadir, id_str+'.pid')
		self.logfile = os.path.join(self.datadir, id_str+'.log')

		self.use_pidfile = sys.platform == 'linux'

		self.proxy = proxy
		self.monerod_addr = monerod_addr
		self.monerod_port = (
			None if monerod_addr else
			CoinDaemon(
				cfg        = self.cfg,
				proto      = proto,
				test_suite = test_suite).rpc_port
		)

		if test_monerod and self.monerod_port:
			import socket
			try:
				socket.create_connection(('localhost', self.monerod_port), timeout=1).close()
			except:
				die('SocketError', f'Unable to connect to Monero daemon at localhost:{self.monerod_port}')

		self.user = user or self.cfg.wallet_rpc_user or self.cfg.monero_wallet_rpc_user
		self.passwd = passwd or self.cfg.wallet_rpc_password or self.cfg.monero_wallet_rpc_password

		assert self.user
		if not self.passwd:
			die(1,
				'You must set your Monero wallet RPC password.\n' +
				'This can be done on the command line with the --wallet-rpc-password option\n' +
				"(insecure, not recommended), or by setting 'monero_wallet_rpc_password' in\n" +
				"the MMGen config file.")

		self.daemon_args = list_gen(
			['--trusted-daemon', trust_monerod],
			['--untrusted-daemon', not trust_monerod],
			[f'--rpc-bind-port={self.rpc_port}'],
			[f'--wallet-dir={self.wallet_dir}'],
			[f'--log-file={self.logfile}'],
			[f'--rpc-login={self.user}:{self.passwd}'],
			[f'--daemon-address={self.monerod_addr}', self.monerod_addr],
			[f'--daemon-port={self.monerod_port}',    not self.monerod_addr],
			[f'--proxy={self.proxy}',                self.proxy],
			[f'--pidfile={self.pidfile}',            self.platform == 'linux'],
			['--detach',                             not (self.opt.no_daemonize or self.platform=='win32')],
			['--stagenet',                           self.network == 'testnet'],
			['--allow-mismatched-daemon-version',    test_suite],
		)

		from .rpc import MoneroWalletRPCClient
		self.rpc = MoneroWalletRPCClient(
			cfg             = self.cfg,
			daemon          = self,
			test_connection = False)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
test: functions required by the MMGen test suite
"""

import sys
from hashlib import sha256
rand_h = sha256('.'.join(sys.argv).encode())

def fake_urandom(n):

	def gen(rounds):
		for _ in range(rounds):
			rand_h.update(b'foo')
			yield rand_h.digest()

	return b''.join(gen(int(n/32)+1))[:n]



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
swap.proto.thorchain: THORChain swap protocol implementation for the MMGen Wallet suite
"""

__all__ = ['SwapCfg', 'SwapAsset', 'Memo']

name = 'THORChain'
exp_prec = 4

from ....util2 import ExpInt
class ExpInt4(ExpInt):
	def __new__(cls, spec):
		return ExpInt.__new__(cls, spec, prec=exp_prec)

def rpc_client(tx, amt):
	from .thornode import Thornode
	return Thornode(tx, amt)

from .cfg import THORChainSwapCfg as SwapCfg

from .asset import THORChainSwapAsset as SwapAsset

from .memo import THORChainMemo as Memo



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
tool.filecrypt: File encryption/decryption routines for the 'mmgen-tool' utility
"""

import os

from .common import tool_cmd_base
from ..crypto import Crypto
from ..fileutil import get_data_from_file, write_data_to_file

class tool_cmd(tool_cmd_base):
	"""
	file encryption and decryption

	MMGen encryption suite:
	* Key: Scrypt (user-configurable hash parameters, 32-byte salt)
	* Enc: AES256_CTR, 16-byte rand IV, sha256 hash + 32-byte nonce + data
	* The encrypted file is indistinguishable from random data
	"""
	def encrypt(self, infile: str, *, outfile='', hash_preset=''):
		"encrypt a file"
		data = get_data_from_file(self.cfg, infile, desc='data for encryption', binary=True)
		enc_d = Crypto(self.cfg).mmgen_encrypt(data, desc='data', hash_preset=hash_preset)
		if not outfile:
			outfile = f'{os.path.basename(infile)}.{Crypto.mmenc_ext}'
		write_data_to_file(self.cfg, outfile, enc_d, desc='encrypted data', binary=True)
		return True

	def decrypt(self, infile: str, *, outfile='', hash_preset=''):
		"decrypt a file"
		enc_d = get_data_from_file(self.cfg, infile, desc='encrypted data', binary=True)
		while True:
			dec_d = Crypto(self.cfg).mmgen_decrypt(enc_d, desc='data', hash_preset=hash_preset)
			if dec_d:
				break
			from ..util import msg
			msg('Trying again...')
		if not outfile:
			from ..util import remove_extension
			o = os.path.basename(infile)
			outfile = remove_extension(o, Crypto.mmenc_ext)
			if outfile == o:
				outfile += '.dec'
		write_data_to_file(self.cfg, outfile, dec_d, desc='decrypted data', binary=True)
		return True



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.eth.tw.json: export and import tracking wallet to JSON format
"""

from collections import namedtuple
from ....tw.json import TwJSON
from ....tw.shared import TwMMGenID

class EthereumTwJSON(TwJSON):

	class Base(TwJSON.Base):

		def __init__(self, proto, *args, **kwargs):

			self.params_keys = ['symbol', 'decimals']
			self.params_tuple = namedtuple('params_tuple', self.params_keys)

			super().__init__(proto, *args, **kwargs)

		@property
		def mappings_json(self):

			def gen_mappings(data):
				for d in data:
					yield (d.mmgen_id, d.address) if hasattr(d, 'mmgen_id') else d

			return self.json_dump({
				'accounts': list(gen_mappings(self.entries['accounts'])),
				'tokens': {k:list(gen_mappings(v)) for k, v in self.entries['tokens'].items()}
			})

		@property
		def num_entries(self):
			return len(self.entries['accounts']) + len(self.entries['tokens'])

	class Import(TwJSON.Import, Base):

		info_msg = """
			This utility will recreate a new tracking wallet from the supplied JSON dump.
			If the dump contains address balances, balances will be updated from it.
		"""

		@property
		async def tracking_wallet_exists(self):
			return bool(self.twctl.data['accounts'] or self.twctl.data['tokens'])

		async def create_tracking_wallet(self):
			return True

		async def get_entries(self):

			edata = self.data['data']['entries']

			def gen_entries(data):
				for d in data:
					if len(d) == 2:
						yield self.params_tuple(*d)
					else:
						e = self.entry_tuple_in(*d)
						yield self.entry_tuple(
							TwMMGenID(self.proto, e.mmgen_id),
							e.address,
							getattr(e, 'amount', '0'),
							e.comment)

			def gen_token_entries():
				for token_addr, token_data in edata['tokens'].items():
					yield (
						token_addr,
						list(gen_entries(token_data)),
					)

			return {
				'accounts': list(gen_entries(edata['accounts'])),
				'tokens': dict(list(gen_token_entries()))
			}

		async def do_import(self, batch):

			from ....obj import TwComment

			def gen_data(data):
				for d in data:
					if hasattr(d, 'address'):
						yield (
							d.address,
							{'mmid': d.mmgen_id, 'comment': TwComment(d.comment)}
							| ({} if d.amount is None else {'balance': d.amount}))
					else:
						yield ('params', {'symbol': d.symbol, 'decimals': d.decimals})

			self.twctl.data = { # keys must be in correct order
				'coin': self.coin.upper(),
				'network': self.network.upper(),
				'accounts': dict(gen_data(self.entries['accounts'])),
				'tokens': {k:dict(gen_data(v)) for k, v in self.entries['tokens'].items()},
			}
			self.twctl.write(quiet=False)

	class Export(TwJSON.Export, Base):

		async def get_entries(self, *, include_amts=True):

			def gen_data(data):
				for k, v in data.items():
					if k == 'params':
						yield self.params_tuple(**v)
					elif include_amts:
						yield self.entry_tuple(TwMMGenID(self.proto, v['mmid']), k, v.get('balance'), v['comment'])
					else:
						yield self.entry_tuple_in(TwMMGenID(self.proto, v['mmid']), k, v['comment'])

			def gen_token_data():
				for token_addr, token_data in self.twctl.data['tokens'].items():
					yield (
						token_addr,
						sorted(
							gen_data(token_data),
							key = lambda x: x.mmgen_id.sort_key if hasattr(x, 'mmgen_id') else '+'
						)
					)

			return {
				'accounts': sorted(
					gen_data(self.twctl.data['accounts']),
					key = lambda x: x.mmgen_id.sort_key),
				'tokens': dict(sorted(gen_token_data()))
			}

		@property
		async def entries_out(self):
			return await self.get_entries(include_amts='amount' in self.keys)

		@property
		async def total(self):
			return sum(self.proto.coin_amt(i.amount) for i in self.entries['accounts']) or self.proto.coin_amt('0')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
color: color handling for the MMGen suite
"""

_colors = {
	'black':   (232,        (30, 0)),
	'red':     (210,        (31, 1)),
	'green':   (121,        (32, 1)),
	'yellow':  (229,        (33, 1)),
	'blue':    (75,         (34, 1)),
	'magenta': (205,        (35, 1)),
	'cyan':    (122,        (36, 1)),

	'gray':    (246,        (30, 1)),
	'orange':  (216,        (31, 1)),
	'purple':  (141,        (35, 1)),
	'pink':    (218,        (35, 1)),

	'melon':   (222,        (33, 1)),
	'brown':   (173,        (33, 0)),
	'grndim':  (108,        (32, 0)),

	'redbg':   ((232, 210), (30, 101)),
	'grnbg':   ((232, 121), (30, 102)),
	'yelbg':   ((232, 229), (30, 103)),
	'blubg':   ((232, 75),  (30, 104)),
}

def nocolor(s):
	return s

def set_vt100():
	'hack to put term into VT100 mode under MSWin'
	import sys
	if sys.platform == 'win32':
		from subprocess import run
		run([], shell=True)

def get_terminfo_colors(term=None):
	from subprocess import run, PIPE
	cmd = ['infocmp', '-0']
	if term:
		cmd.append(term)

	try:
		cmdout = run(cmd, stdout=PIPE, check=True).stdout.decode()
	except:
		set_vt100()
		return None
	else:
		set_vt100()
		s = [e.split('#')[1] for e in cmdout.split(',') if e.startswith('colors')][0]
		from .util import is_hex_str
		if s.isdecimal():
			return int(s)
		elif s.startswith('0x') and is_hex_str(s[2:]):
			return int(s[2:], 16)
		else:
			return None

def init_color(num_colors='auto'):
	assert num_colors in ('auto', 8, 16, 256, 0)

	import sys
	self = sys.modules[__name__]

	if num_colors == 'auto':
		import os
		if sys.platform == 'win32':
			# Force 256-color for MSYS2: terminal supports it, however infocmp reports 8-color.
			# We also avoid spawning a subprocess, leading to a subsequent OSError 22 when testing
			# with pexpect spawn.
			num_colors = 256
		elif (t := os.getenv('TERM')) and t.endswith('256color'):
			num_colors = 256
		else:
			num_colors = get_terminfo_colors() or 16

	reset = '\033[0m'
	if num_colors == 0:
		ncc = (lambda s: s).__code__
		for c in _colors:
			getattr(self, c).__code__ = ncc
	elif num_colors == 256:
		for c, e in _colors.items():
			start = (
				'\033[38;5;{};1m'.format(e[0]) if type(e[0]) == int else
				'\033[38;5;{};48;5;{};1m'.format(*e[0]))
			getattr(self, c).__code__ = eval(f'(lambda s: "{start}" + s + "{reset}").__code__')
	elif num_colors in (8, 16):
		for c, e in _colors.items():
			start = (
				'\033[{}m'.format(e[1][0]) if e[1][1] == 0 else
				'\033[{};{}m'.format(*e[1]))
			getattr(self, c).__code__ = eval(f'(lambda s: "{start}" + s + "{reset}").__code__')

	set_vt100()

# Each color name must be bound to an independent stub function with its own
# address in memory.  The names themselves must never be redefined, since other
# modules could import them before init_color() is run.  Instead, the code
# objects of their associated functions are manipulated by init_color() to
# enable/disable color.

black   = lambda s: s
red     = lambda s: s
green   = lambda s: s
yellow  = lambda s: s
blue    = lambda s: s
magenta = lambda s: s
cyan    = lambda s: s

gray    = lambda s: s
orange  = lambda s: s
purple  = lambda s: s
pink    = lambda s: s

melon   = lambda s: s
brown   = lambda s: s
grndim  = lambda s: s

redbg   = lambda s: s
grnbg   = lambda s: s
yelbg   = lambda s: s
blubg   = lambda s: s



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.rune.params: THORChain protocol
"""

from ...protocol import CoinProtocol, decoded_addr, _nw
from ...obj import Hostname
from ...addr import CoinAddr
from ...contrib import bech32

from ..btc.params import mainnet as btc_mainnet

class mainnet(CoinProtocol.Secp256k1):
	mod_clsname     = 'THORChain'
	network_names   = _nw('mainnet', 'stagenet', 'testnet')
	chain_id        = 'thorchain-1'
	mmtypes         = ('X',)
	preferred_mmtypes  = ('X',)
	dfl_mmtype      = 'X'
	coin_amt        = 'UniAmt'
	max_tx_fee      = 0.1
	caps            = ()
	mmcaps          = ('tw', 'rpc', 'rpc_init')
	base_proto      = 'THORChain'
	base_proto_coin = 'RUNE'
	base_coin       = 'RUNE'
	bech32_hrp      = 'thor'
	sign_mode       = 'standalone'
	rpc_type        = 'remote'
	avg_bdi         = 6 # TODO
	has_usr_fee     = False
	is_vm           = True
	address_reuse_ok = False

	wif_ver_num = btc_mainnet.wif_ver_num
	coin_cfg_opts = btc_mainnet.coin_cfg_opts
	encode_wif = btc_mainnet.encode_wif
	decode_wif = btc_mainnet.decode_wif

	rpc_remote_params      = {'server_domain': Hostname('ninerealms.com')}
	rpc_remote_rest_params = {'host': Hostname('thornode.ninerealms.com')}
	rpc_remote_rpc_params  = {'host': Hostname('rpc.ninerealms.com')}
	rpc_swap_params        = {'host': Hostname('thornode.ninerealms.com')}

	def decode_addr(self, addr):
		hrp, data = bech32.bech32_decode(addr)
		assert hrp == self.bech32_hrp, f'{hrp!r}: invalid bech32 hrp (should be {self.bech32_hrp!r})'
		return decoded_addr(
			bytes(bech32.convertbits(data, 5, 8)),
			None,
			'bech32') if data else False

	def encode_addr_bech32x(self, pubhash):
		return CoinAddr(
			self,
			bech32.bech32_encode(
				hrp  = self.bech32_hrp,
				data = bech32.convertbits(list(pubhash), 8, 5)))

class testnet(mainnet): # testnet is stagenet
	bech32_hrp = 'sthor'
	rpc_remote_rest_params = {'host': Hostname('stagenet-thornode.ninerealms.com')}
	rpc_remote_rpc_params  = {'host': Hostname('stagenet-rpc.ninerealms.com')}
	rpc_swap_params        = {'host': Hostname('stagenet-thornode.ninerealms.com')}

class regtest(testnet): # regtest is deprecated testnet
	bech32_hrp = 'tthor'
	rpc_remote_params = {
		'server_domain': Hostname('localhost')}
	rpc_remote_rest_params = {
		'network_proto': 'http',
		'host': Hostname('localhost:18800'),
		'verify': False}
	rpc_remote_rpc_params = rpc_remote_rest_params
	rpc_swap_params = rpc_remote_rest_params | {'host': Hostname('localhost:18900')}



import os as overlay_fake_os
from .crypto_orig import *

if overlay_fake_os.getenv('MMGEN_TEST_SUITE_DETERMINISTIC'):

	overlay_fake_get_random_orig = Crypto.get_random
	overlay_fake_add_user_random_orig = Crypto.add_user_random

	from .test import fake_urandom as overlay_fake_urandom

	Crypto.get_random = lambda self, length: overlay_fake_urandom(
		len(overlay_fake_get_random_orig(self, length)))

	Crypto.add_user_random = lambda self, rand_bytes, desc: overlay_fake_urandom(
		len(overlay_fake_add_user_random_orig(self, rand_bytes, desc=desc)))



#!/usr/bin/env python3

"""
test.modtest_d.ecc: elliptic curve unit test for the MMGen suite
"""

from mmgen.proto.secp256k1.secp256k1 import (
	pubkey_gen,
	pubkey_tweak_add,
	pubkey_check,
	sign_msghash,
	pubkey_recover,
	verify_sig)

from ..include.common import vmsg
from ..include.ecc import pubkey_tweak_add_pyecdsa, sign_msghash_pyecdsa, verify_sig_pyecdsa
from mmgen.protocol import CoinProtocol

secp256k1_group_order = CoinProtocol.Secp256k1.secp256k1_group_order

class unit_tests:

	def sig_ops(self, name, ut):
		vmsg('  Creating and verifying signatures and recovering public keys:')
		for mh, pk in (
				(1,                  1),
				(123456789 * 2**222, 12345),
				(123456789 * 2**222, 2**256 - 2**129 - 987654321),
				(9999999,            2**233),
				(12345,              1234 * 2**240),
			):
			msghash = mh.to_bytes(32, 'big')
			privkey = pk.to_bytes(32, 'big')
			vmsg(f'\n   msg:     {msghash.hex()}')
			vmsg(f'   privkey: {privkey.hex()}')
			pubkey = pubkey_gen(privkey, 1)
			sig, recid = sign_msghash(msghash, privkey)
			sig_chk = sign_msghash_pyecdsa(msghash, privkey)
			if sig != sig_chk:
				import time
				from mmgen.util import ymsg
				ymsg(f'Warning: signature ({sig.hex()}) doesn’t match reference value ({sig_chk.hex()})!')
				time.sleep(1)
			vmsg(f'   recid:   {recid}')
			assert recid in (0, 1)
			assert verify_sig(sig, msghash, pubkey) == 1, 'signature verification failed (secp256k1)'
			assert verify_sig_pyecdsa(sig, msghash, pubkey) == 1, 'signature verification failed (ecdsa)'
			pubkey_rec = pubkey_recover(msghash, sig, recid, True)
			assert pubkey == pubkey_rec, f'{pubkey.hex()} != {pubkey_rec.hex()}'
		return True

	def sig_errors(self, name, ut):
		vmsg('  Testing error handling for signature ops')

		msghash = bytes.fromhex('deadbeef' * 8)
		privkey = bytes.fromhex('beadcafe' * 8)
		pubkey = pubkey_gen(privkey, 1)
		sig, recid = sign_msghash(msghash, privkey)

		def sign1(): sign_msghash(1, bytes(32))
		def sign2(): sign_msghash(b'\xff' + bytes(32), bytes(32))
		def sign3(): sign_msghash(bytes(32), bytes(32))

		def verify1(): verify_sig(1, 2, 3)
		def verify2(): verify_sig(bytes(64), bytes(32), bytes(33))
		def verify3(): assert verify_sig(bytes([99]) + sig[1:], msghash, pubkey) == 1, 'bad signature'
		def verify4(): assert verify_sig(sig, msghash, pubkey) == 0, 'good signature'
		def verify5(): verify_sig(sig, msghash + bytes([201]), pubkey)
		def verify6(): verify_sig(sig + bytes([66]), msghash, pubkey)

		def recov1(): pubkey_recover(1, 2, 3)
		def recov2(): pubkey_recover(msghash, sig, 8, True)
		def recov3(): pubkey_recover(msghash, sig, -3, 1)
		def recov4(): pubkey_recover(msghash, bytes([77]) + sig, recid, 1)
		def recov5(): pubkey_recover(msghash + bytes([33]), sig, recid, 1)
		def recov6():
			assert pubkey_recover(msghash[:-1] + bytes([44]), sig, recid, 1) == pubkey, 'bad pubkey'
		def recov7():
			assert pubkey_recover(msghash, sig, recid, True) != pubkey, 'good pubkey'

		bad_data = (
			('sign: bad args',           'ValueError',     'Unable to parse',              sign1),
			('sign: bad msghash len',    'RuntimeError',   'hash length',                  sign2),
			('sign: privkey=0',          'ValueError',     'Private key not in allowable', sign3),
			('verify: bad args',         'ValueError',     'Unable to parse',              verify1),
			('verify: bad pubkey',       'RuntimeError',   'Failed to parse',              verify2),
			('verify: bad sig',          'AssertionError', 'bad signature',                verify3),
			('verify: good sig',         'AssertionError', 'good signature',               verify4),
			('verify: bad msghash len',  'RuntimeError',   'message hash length',          verify5),
			('verify: bad sig len',      'RuntimeError',   'Invalid signature length',     verify6),
			('recover: bad args',        'ValueError',     'Unable to parse',              recov1),
			('recover: bad recid',       'RuntimeError',   'Invalid recovery ID',          recov2),
			('recover: bad recid',       'RuntimeError',   'Invalid recovery ID',          recov3),
			('recover: bad sig len',     'RuntimeError',   'Invalid signature length',     recov4),
			('recover: bad msghash len', 'RuntimeError',   'message hash length',          recov5),
			('recover: bad pubkey',      'AssertionError', 'bad pubkey',                   recov6),
			('recover: bad pubkey',      'AssertionError', 'good pubkey',                  recov7),
		)

		ut.process_bad_data(bad_data, pfx='')
		return True

	def pubkey_ops(self, name, ut):
		vmsg('  Generating pubkey, adding scalar 123456789 to pubkey:')
		pk_addend_bytes = int.to_bytes(123456789, length=32, byteorder='big')

		for privkey in (
				'beadcafe' * 8,
				f'{1:064x}',
				f'{secp256k1_group_order-1:x}',
			):
			vmsg(f'  privkey = 0x{privkey}')
			for compressed, length in ((False, 65), (True, 33)):
				vmsg(f'    {compressed=}')
				pubkey_bytes = pubkey_gen(bytes.fromhex(privkey), int(compressed))
				pubkey_check(pubkey_bytes)
				vmsg(f'      pubkey:  {pubkey_bytes.hex()}')

				res1 = pubkey_tweak_add(pubkey_bytes, pk_addend_bytes)
				pubkey_check(res1)
				vmsg(f'      tweaked: {res1.hex()}')

				res2 = pubkey_tweak_add_pyecdsa(pubkey_bytes, pk_addend_bytes)
				pubkey_check(res2)

				assert len(res1) == length
				assert res1 == res2

		return True

	def pubkey_errors(self, name, ut):
		vmsg('  Testing error handling for public key ops')

		def gen1(): pubkey_gen(bytes(32), 1)
		def gen2(): pubkey_gen(secp256k1_group_order.to_bytes(length=32, byteorder='big'), 1)
		def gen3(): pubkey_gen((secp256k1_group_order+1).to_bytes(length=32, byteorder='big'), 1)
		def gen4(): pubkey_gen(bytes.fromhex('ff'*32), 1)
		def gen5(): pubkey_gen(bytes.fromhex('ab'*31), 1)
		def gen6(): pubkey_gen(bytes.fromhex('ab'*33), 1)

		pubkey_bytes = pubkey_gen(bytes.fromhex('beadcafe'*8), 1)
		def tweak1(): pubkey_tweak_add(pubkey_bytes, bytes(32))
		def tweak2(): pubkey_tweak_add(bytes.fromhex('03'*64), int.to_bytes(1, length=32, byteorder='big'))

		def check1(): pubkey_check(bytes.fromhex('04'*33))
		def check2(): pubkey_check(bytes.fromhex('03'*65))
		def check3(): pubkey_check(bytes.fromhex('02'*65))
		def check4(): pubkey_check(bytes.fromhex('03'*64))
		def check5(): pubkey_check(b'')

		bad_data = (
			('privkey == 0',              'ValueError', 'Private key not in allowable range', gen1),
			('privkey == group order',    'ValueError', 'Private key not in allowable range', gen2),
			('privkey == group order+1',  'ValueError', 'Private key not in allowable range', gen3),
			('privkey == 2^256-1',        'ValueError', 'Private key not in allowable range', gen4),
			('len(privkey) == 31',        'ValueError', 'Private key length not 32 bytes',    gen5),
			('len(privkey) == 33',        'ValueError', 'Private key length not 32 bytes',    gen6),

			('tweak == 0',                'ValueError', 'Tweak not in allowable range',       tweak1),
			('pubkey length == 64',       'ValueError', 'Serialized public key length not',   tweak2),

			('invalid pubkey (33 bytes)', 'ValueError', 'Invalid first byte',                 check1),
			('invalid pubkey (65 bytes)', 'ValueError', 'Invalid first byte',                 check2),
			('invalid pubkey (65 bytes)', 'ValueError', 'Invalid first byte',                 check3),
			('pubkey length == 64',       'ValueError', 'Serialized public key length not',   check4),
			('pubkey length == 0',        'ValueError', 'Serialized public key length not',   check5),
		)

		ut.process_bad_data(bad_data, pfx='')
		return True



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
help.swaptxcreate_examples: swaptxcreate and swaptxdo help examples for the MMGen Wallet suite
"""

from ..cfg import gc

def help(proto, cfg):

	return f"""
EXAMPLES:

  Display available swap assets:

    $ {gc.prog_name} -S

  Create a BTC-to-LTC swap transaction, prompting the user for transaction
  inputs.  The full value of the inputs, minus miner fees, will be swapped
  and sent to an unused address in the user’s LTC tracking wallet:

    $ {gc.prog_name} BTC LTC

  Same as above, but swap 0.123 BTC, minus miner fees, and send the change to
  an unused address in the BTC tracking wallet:

    $ {gc.prog_name} BTC 0.123 LTC

  Same as above, but specify that the change address be a Segwit P2SH (‘S’)
  address:

    $ {gc.prog_name} BTC 0.123 S LTC

  Same as above, but additionally specify that the destination LTC address be
  a compressed P2PKH (‘C’) address:

    $ {gc.prog_name} BTC 0.123 S LTC C

  Same as above, but specify the BTC change address explicitly and the
  destination LTC address by Seed ID and address type:

    $ {gc.prog_name} BTC 0.123 BEADCAFE:S:6 LTC BEADCAFE:C

  Abort the above swap by creating a replacement transaction that returns the
  funds to the originating tracking wallet (omit the transaction filename if
  using --autosign):

    $ mmgen-txbump BEADCAFE:S:6 [raw transaction file]

  Swap 0.123 BTC to a non-wallet address (not recommended):

    $ {gc.prog_name} BTC 0.123 LTC ltc1qaq8t3pakcftpk095tnqfv5cmmczysls0xx9388

  Create an LTC-to-BCH swap transaction, with the Litecoin daemon running on
  host ‘orion’ and Bitcoin Cash Node daemon on host ‘gemini’ with non-standard
  RPC port 8332.  Communicate with the swap quote server via Tor.

    $ {gc.prog_name} --ltc-rpc-host=orion --bch-rpc-host=gemini --bch-rpc-port=8332 --proxy=localhost:9050 LTC BCH

  After sending, check the status of the above swap’s LTC deposit transaction
  (omit the transaction filename if using --autosign):

    $ mmgen-txsend --ltc-rpc-host=orion --status [transaction file]

  Check whether the funds have arrived in the BCH destination wallet:

    $ mmgen-tool --coin=bch --bch-rpc-host=gemini twview minconf=0

  Create a Tether-to-LTC swap transaction for autosigning, connecting to the
  swap quote server via Tor:

    $ {gc.prog_name} --autosign --proxy=localhost:9050 ETH.USDT 1000 LTC

  After signing, send the transaction via public Etherscan proxy over Tor:

    $ mmgen-txsend --autosign --quiet --tx-proxy=etherscan --proxy=localhost:9050

  After sending, check the transaction status:

    $ mmgen-txsend --autosign --verbose --status

  Create a Tether-to-DAI swap transaction, with explicit destination account:

    $ {gc.prog_name} ETH.USDT 1000 ETH.DAI E:01234ABC:3

  Create a RUNE-to-BTC swap transaction, proxying requests to the remote
  Thornode server via Tor:

    $ {gc.prog_name} --proxy=localhost:9050 RUNE 1000 BTC

  Same as above, but proxy requests via the I2P router running on host gw1:

    $ https_proxy=http://gw1:4444 {gc.prog_name} --proxy=env RUNE 1000 BTC
"""



#!/usr/bin/env python3

"""
test.modtest_d.devtools: devtools unit tests for the MMGen suite
"""

import os, json
from mmgen.util import msg
from mmgen.devtools import print_diff, get_ndiff, print_stack_trace, pmsg_r, pmsg, Pmsg

textA = """
def main():
	a = 1
	b = 2
	c = 3
""".lstrip()

textB = """
def main():
	a = 1
	b = 0
	c = 3
""".lstrip()

with open('test/ref/ethereum/tracking-wallet-v1.json') as fh:
	jsonA = fh.read()
dataB = json.loads(jsonA)
dataB['coin'] = 'ETC'
jsonB = json.dumps(dataB)

text_data = (
	(textA, textB, 'a/main.py',   'b/main.py',   False, 'text: one line difference'),
	('',    textB, 'a/main.py',   'b/main.py',   False, 'text: first file empty'),
	(textA, textA, 'a/main.py',   'b/main.py',   False, 'text: identical files'),
	('',    '',    'a/empty.txt', 'b/empty.txt', False, 'text: empty files'),
)

json_data = (
	(jsonA, jsonB, 'a/data.json', 'b/data.json', True,  'json: one difference'),
	('{}',  jsonB, 'a/data.json', 'b/data.json', True,  'json: first file empty'),
	(jsonA, jsonA, 'a/data.json', 'b/data.json', True,  'json: identical files'),
	('{}',  '{}',  'a/data.json', 'b/data.json', True,  'json: empty files'),
)

def print_hdr(hdr):
	print('{a} {b} {c}'.format(
		a = '-' * ((78 - len(hdr))//2),
		b = hdr,
		c = '-' * ((78 - len(hdr))//2 + (len(hdr) % 2))))

# TODO: add data checks
class unit_tests:

	silence_output = True

	def _post_subtest(self, name, subname, ut):
		print('-' * 80 + '\n')

	def diff(self, name, ut):
		for data in text_data + json_data:
			print_hdr(data[-1])
			print_diff(*data[:-1])
		return True

	def ndiff(self, name, ut):
		for data in text_data:
			print_hdr(data[-1])
			print('\n'.join(get_ndiff(*data[:2])))
		return True

	def stack_trace(self, name, ut):
		print_hdr('stack trace')
		print_stack_trace('Test', fh_list=[open(os.devnull, 'w')], trim=0)
		return True

	def obj_pmsg(self, name, ut):
		from mmgen.protocol import init_proto
		from mmgen.seed import Seed
		from mmgen.addrlist import AddrList
		from ..include.common import cfg
		print_hdr('MMGenObject.pmsg()')
		AddrList(
			cfg         = cfg,
			proto       = init_proto(cfg, 'btc'),
			seed        = Seed(cfg, seed_bin=bytes.fromhex('bead'*16)),
			addr_idxs   = '1',
			mmtype      = 'B',
			skip_chksum = True).pmsg(color='green')
		return True

	def pmsg(self, name, ut):
		colors = (None, 'red', 'green', 'yellow', 'blue', 'purple')

		msg('\npmsg_r():')
		for color in colors:
			pmsg_r({'color':color}, color=color)

		msg('\n\npmsg():')
		for color in colors:
			pmsg({'color':color}, color=color)

		msg('\nPmsg():')
		for color in colors:
			Pmsg({'color':color}, color=color)

		return True



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.vm.tx.completed: completed transaction methods for VM chains
"""

class Completed:

	def get_swap_memo_maybe(self):
		return self.swap_memo.encode() if getattr(self, 'swap_memo', None) else None

	@property
	def send_amt(self):
		return self.outputs[0].amt if self.outputs else self.proto.coin_amt('0')

	@property
	def change(self):
		return self.sum_inputs() - self.send_amt - self.fee

	def check_txfile_hex_data(self):
		pass

	def check_sigs(self): # TODO
		from ....util import is_hex_str
		if is_hex_str(self.serialized):
			return True
		return False

	def check_pubkey_scripts(self):
		pass

	def get_serialized_locktime(self):
		return None # TODO



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
wallet.mnemonic: MMGen mnemonic wallet base class
"""

from ..baseconv import baseconv
from ..util import msg
from ..seed import Seed
from .unenc import wallet

class wallet(wallet):

	stdin_ok = True
	conv_cls = baseconv
	choose_seedlen_prompt = 'Choose a mnemonic length: 1) 12 words, 2) 18 words, 3) 24 words: '
	choose_seedlen_confirm = 'Mnemonic length of {} words chosen. OK?'

	@property
	def mn_lens(self):
		return sorted(self.conv_cls(self.wl_id).seedlen_map_rev)

	def _get_data_from_user(self, desc):

		if not self.cfg.stdin_tty:
			from ..ui import get_data_from_user
			return get_data_from_user(self.cfg, desc=desc)

		self._print_seed_type()

		if self.cfg.seed_len:
			from ..obj import Int
			msg('Using seed length {} (user-configured)'.format(Int(self.cfg.seed_len).hl()))
			mn_len = self.conv_cls(self.wl_id).seedlen_map[self.cfg.seed_len // 8]
		else:
			mn_len = self._choose_seedlen(self.mn_lens)

		from ..mn_entry import mn_entry
		self.mnemonic = mn_entry(self.cfg, self.wl_id)

		return self.mnemonic.get_mnemonic_from_user(mn_len)

	def _format(self):

		seed = self.seed.data

		bc = self.conv_cls(self.wl_id)
		mn  = bc.frombytes(seed, pad='seed')
		rev = bc.tobytes(mn, pad='seed')

		# Internal error, so just die on fail
		self.cfg._util.compare_or_die(rev, 'recomputed seed', seed, 'original seed', e='Internal error')

		self.ssdata.mnemonic = mn
		self.fmt_data = ' '.join(mn) + '\n'

	def _deformat(self):

		bc = self.conv_cls(self.wl_id)
		mn = self.fmt_data.split()

		if len(mn) not in self.mn_lens:
			msg('Invalid mnemonic ({} words).  Valid numbers of words: {}'.format(
				len(mn),
				', '.join(map(str, self.mn_lens))))
			return False

		for n, w in enumerate(mn, 1):
			if w not in bc.digits:
				msg(f'Invalid mnemonic: word #{n} is not in the {self.wl_id.upper()} wordlist')
				return False

		seed = bc.tobytes(mn, pad='seed')
		rev  = bc.frombytes(seed, pad='seed')

		if len(seed) * 8 not in Seed.lens:
			msg('Invalid mnemonic (produces too large a number)')
			return False

		# Internal error, so just die:
		self.cfg._util.compare_or_die(
			val1  = ' '.join(rev),
			val2  = ' '.join(mn),
			desc1 = 'recomputed mnemonic',
			desc2 = 'original mnemonic',
			e     = 'Internal error')

		self.seed = Seed(self.cfg, seed_bin=seed)
		self.ssdata.mnemonic = mn

		self.check_usr_seed_len()

		return True



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.eth.tx.base: Ethereum base transaction class
"""

from collections import namedtuple

from ....tx.base import Base as TxBase
from ....obj import Int

class Base(TxBase):

	rel_fee_desc = 'gas price'
	rel_fee_disp = 'gas price in Gwei'
	txobj = None
	dfl_gas = 21000 # the startGas amt used in the transaction
	                # for simple sends with no data, startGas = 21000
	contract_desc = 'contract'
	usr_contract_data = b''
	disable_fee_check = False

	@property
	def nondata_outputs(self):
		return self.outputs

	def pretty_fmt_fee(self, fee):
		if fee < 10:
			return f'{fee:.3f}'.rstrip('0').rstrip('.')
		return str(int(fee))

	# given absolute fee in ETH, return gas price for display in selected unit
	def fee_abs2rel(self, abs_fee, *, to_unit='Gwei'):
		return self.pretty_fmt_fee(
			self.fee_abs2gasprice(abs_fee).to_unit(to_unit))

	# given absolute fee in ETH, return gas price in ETH
	def fee_abs2gasprice(self, abs_fee):
		return self.proto.coin_amt(int(abs_fee.toWei() // self.total_gas), from_unit='wei')

	# given rel fee (gasPrice) in wei, return absolute fee using self.total_gas
	def fee_gasPrice2abs(self, rel_fee):
		assert isinstance(rel_fee, int), f'{rel_fee!r}: incorrect type for fee estimate (not an integer)'
		return self.proto.coin_amt(rel_fee * self.total_gas, from_unit='wei')

	def is_replaceable(self):
		return True

	async def get_receipt(self, txid, *, receipt_only=False):
		import asyncio
		from ....util import msg, msg_r

		for n in range(60):
			rx = await self.rpc.call('eth_getTransactionReceipt', '0x'+txid) # -> null if pending
			if rx or not self.cfg.wait:
				break
			if n == 0:
				msg_r('Waiting for first confirmation..')
			await asyncio.sleep(1)
			msg_r('.')

		if rx:
			if n:
				msg('OK')
			if receipt_only:
				return rx
		else:
			if self.cfg.wait:
				msg('timeout exceeded!')
			return None

		tx = await self.rpc.call('eth_getTransactionByHash', '0x'+txid)
		return namedtuple('exec_status',
				['status', 'gas_sent', 'gas_used', 'gas_price', 'contract_addr', 'tx', 'rx'])(
			status        = Int(rx['status'], base=16), # zero is failure, non-zero success
			gas_sent      = Int(tx['gas'], base=16),
			gas_used      = Int(rx['gasUsed'], base=16),
			gas_price     = self.proto.coin_amt(Int(tx['gasPrice'], base=16), from_unit='wei'),
			contract_addr = self.proto.coin_addr(rx['contractAddress'][2:])
				if rx['contractAddress'] else None,
			tx = tx,
			rx = rx)

	def check_serialized_integrity(self):
		if self.signed:
			from .. import rlp
			o = self.txobj
			d = rlp.decode(bytes.fromhex(self.serialized))
			to_key = 'token_addr' if self.is_token else 'to'

			if o['nonce'] == 0:
				assert d[0] == b'', f'{d[0]}: invalid nonce in serialized data'
			else:
				assert int(d[0].hex(), 16) == o['nonce'], f'{d[0]}: invalid nonce in serialized data'
			if o.get(to_key):
				assert d[3].hex() == o[to_key], f'{d[3].hex()}: invalid ‘to’ address in serialized data'
			if not self.is_token:
				if o['amt']:
					assert int(d[4].hex(), 16) == o['amt'].toWei(), (
						f'{d[4].hex()}: invalid amt in serialized data')
				if self.is_swap:
					assert d[5] == self.swap_memo.encode(), (
						f'{d[5]}: invalid swap memo in serialized data')

class TokenBase(Base):
	dfl_gas = 75000
	dfl_router_gas = 150000
	contract_desc = 'token contract'

	def check_serialized_integrity(self):
		if self.signed:
			super().check_serialized_integrity()

			from .. import rlp
			from ....amt import TokenAmt
			d = rlp.decode(bytes.fromhex(self.serialized))
			o = self.txobj

			assert d[4] == b'', f'{d[4]}: non-empty amount field in token transaction in serialized data'

			data = d[5].hex()
			assert data[:8] == ('095ea7b3' if self.is_swap else 'a9059cbb'), (
				f'{data[:8]}: invalid MethodID for op ‘{self.token_op}’ in serialized data')
			assert data[32:72] == o['token_to'], (
				f'{data[32:72]}: invalid ‘token_to‘ address in serialized data')
			assert TokenAmt(
					int(data[72:], 16),
					decimals = o['decimals'],
					from_unit = 'atomic') == o['amt'], (
				f'{data[72:]}: invalid amt in serialized data')

			if self.is_swap:
				d = rlp.decode(bytes.fromhex(self.serialized2))
				data = d[5].hex()
				assert data[:8] == '44bc937b', (
					f'{data[:8]}: invalid MethodID in router TX serialized data')
				assert data[32:72] == self.token_vault_addr, (
					f'{data[32:72]}: invalid vault address in router TX serialized data')

				memo = bytes.fromhex(data[392:])[:len(self.swap_memo)]
				assert memo == self.swap_memo.encode(), (
					f'{memo}: invalid swap memo in router TX serialized data')
				assert TokenAmt(
						int(data[136:200], 16),
						decimals = o['decimals'],
						from_unit = 'atomic') == o['amt'], (
					f'{data[136:200]}: invalid amt in router TX serialized data')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
mmgen-txcreate: Create a cryptocoin transaction with MMGen- and/or non-MMGen
                inputs and outputs
"""

from .cfg import gc, Config
from .util import Msg, fmt_list, async_run

target = gc.prog_name.split('-')[1].removesuffix('create')

opts_data = {
	'filter_codes': {
		'tx':     ['-', 't'],
		'swaptx': ['-', 's'],
	}[target],
	'sets': [('yes', True, 'quiet', True)],
	'text': {
		'desc': {
			'tx':     f'Create a transaction with outputs to specified coin or {gc.proj_name} addresses',
			'swaptx': f'Create a DEX swap transaction from one {gc.proj_name} tracking wallet to another',
		}[target],
		'usage':   '[opts] {u_args} [addr file ...]',
		'options': """
			-- -h, --help            Print this help message
			-- --, --longhelp        Print help message for long (global) options
			-- -a, --autosign        Create a transaction for offline autosigning (see
			+                        ‘mmgen-autosign’). The removable device is mounted and
			+                        unmounted automatically
			r- -A, --fee-adjust=  f  Adjust transaction fee by factor ‘f’ (see below)
			-- -B, --no-blank        Don't blank screen before displaying {a_info}
			-- -c, --comment-file=f  Source the transaction's comment from file 'f'
			b- -C, --fee-estimate-confs=c Desired number of confirmations for fee estimation
			+                        (default: {cfg.fee_estimate_confs})
			-- -d, --outdir=      d  Specify an alternate directory 'd' for output
			e- -D, --contract-data=D Path to file containing hex-encoded contract data
			b- -E, --fee-estimate-mode=M Specify the network fee estimate mode.  Choices:
			+                        {fe_all}.  Default: {fe_dfl!r}
			r- -f, --fee=         f  Transaction fee, as a decimal {cu} amount or as
			+                        {fu} (an integer followed by {fl}).
			+                        See FEE SPECIFICATION below.  If omitted, fee will be
			+                        calculated using network fee estimation.
			et -g, --gas=N           Set the gas limit (see GAS LIMIT below)
			-s -g, --gas=N           Set the gas limit for Ethereum (see GAS LIMIT below)
			-s -G, --router-gas=N    Set the gas limit for the Ethereum router contract
			+                        (integer).  When unset, a hardcoded default will be
			+                        used.  Applicable only for swaps from token assets.
			-- -i, --info            Display {a_info} and exit
			-- -I, --inputs=      i  Specify transaction inputs (comma-separated list of
			+                        MMGen IDs or coin addresses).  Note that ALL unspent
			+                        outputs associated with each address will be included.
			bt -l, --locktime=    t  Lock time (block height or unix seconds) (default: 0)
			-s -l, --trade-limit=L   Minimum swap amount, as either percentage or absolute
			+                        coin amount (see TRADE LIMIT below)
			b- -L, --autochg-ignore-labels Ignore labels when autoselecting change addresses
			-- -m, --minconf=     n  Minimum number of confirmations required to spend
			+                        outputs (default: 1)
			-- -q, --quiet           Suppress warnings; overwrite files without prompting
			-s -r, --stream-interval=N Set block interval for streaming swap (default: {si})
			bt -R, --no-rbf          Make transaction non-replaceable (non-replace-by-fee
			+                        according to BIP 125)
			-s -s, --swap-proto      Swap protocol to use (Default: {x_dfl},
			+                        Choices: {x_all})
			-s -S, --list-assets     List available swap assets
			-- -v, --verbose         Produce more verbose output
			b- -V, --vsize-adj=   f  Adjust transaction's estimated vsize by factor 'f'
			rs -x, --proxy=P         Fetch the swap quote via SOCKS5h proxy ‘P’ (host:port).
			+                        Use special value ‘env’ to honor *_PROXY environment
			+                        vars instead.
			X- -x, --proxy=P         Connect to remote server(s) via SOCKS5h proxy ‘P’
			+                        (host:port).  Use special value ‘env’ to honor *_PROXY
			+                        environment vars instead.
			-- -y, --yes             Answer 'yes' to prompts, suppress non-essential output
			e- -X, --cached-balances Use cached balances
		""",
		'notes': '\n{c}\n{n_at}\n\n{g}{F}{x}',
	},
	'code': {
		'usage': lambda cfg, proto, help_notes, s: s.format(
			u_args = help_notes(f'{target}create_args')),
		'options': lambda cfg, proto, help_notes, s: s.format(
			cfg    = cfg,
			cu     = proto.coin,
			a_info = help_notes('account_info_desc'),
			fu     = help_notes('rel_fee_desc'),
			fl     = help_notes('fee_spec_letters', use_quotes=True),
			si     = help_notes('stream_interval'),
			fe_all = fmt_list(cfg._autoset_opts['fee_estimate_mode'].choices, fmt='no_spc'),
			fe_dfl = cfg._autoset_opts['fee_estimate_mode'].choices[0],
			x_all = fmt_list(cfg._autoset_opts['swap_proto'].choices, fmt='no_spc'),
			x_dfl = cfg._autoset_opts['swap_proto'].choices[0]),
		'notes': lambda cfg, help_mod, help_notes, s: s.format(
			c      = help_mod(f'{target}create'),
			g      = help_notes('gas_limit', target),
			F      = help_notes('fee', all_coins={'tx': False, 'swaptx': True}[target]),
			n_at   = help_notes('address_types'),
			x      = help_mod(f'{target}create_examples'))
	}
}

cfg = Config(opts_data=opts_data)

if cfg.list_assets:
	import sys
	from .tx.new_swap import get_swap_proto_mod
	sp = get_swap_proto_mod(cfg.swap_proto)
	Msg('AVAILABLE SWAP ASSETS:\n' + sp.SwapAsset('BTC', 'send').fmt_assets_data(indent='  '))
	sys.exit(0)

if not (cfg.info or cfg.contract_data) and len(cfg._args) < {'tx': 1, 'swaptx': 2}[target]:
	cfg._usage()

async def main():

	if cfg.autosign:
		from .tx.util import mount_removable_device
		from .autosign import Signable
		asi = mount_removable_device(cfg)
		Signable.automount_transaction(asi).check_create_ok()

	if target == 'swaptx':
		from .tx.new_swap import get_send_proto
		proto = get_send_proto(cfg)
	else:
		proto = cfg._proto

	from .tx import NewTX
	tx1 = NewTX(cfg=cfg, proto=proto, target=target)

	tx2 = await tx1.create(
		cmd_args = cfg._args,
		locktime = int(cfg.locktime or 0),
		do_info  = cfg.info)

	tx2.file.write(
		outdir                = asi.txauto_dir if cfg.autosign else None,
		ask_write             = not cfg.yes,
		ask_overwrite         = not cfg.yes,
		ask_write_default_yes = False)

async_run(main())



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
examples/coin-daemon-info.py:
    Get info about multiple running coin daemons.
    Demonstrates use of the MMGen Config API.
"""

# Instructions
#
#  Testing mode:
#
#   1) From the MMGen repository root, start the mainnet test suite daemons as follows
#      (note that Geth is the default testing daemon for ETH):
#
#       test/start-coin-daemons.py btc ltc eth
#
#   2) Then run the script as follows:
#
#       PYTHONPATH=. MMGEN_TEST_SUITE=1 examples/coin-daemon-info.py btc ltc eth
#
#  Live mode:
#
#   1) Start up one or more of bitcoind, litecoind or geth with the standard mainnet ports
#      and datadirs.  For geth, the options ‘--http --http.api=eth,web3’ are required.
#
#   2) Then run the script as follows:
#
#       PYTHONPATH=. examples/coin-daemon-info.py btc ltc eth

import sys, os, asyncio

from mmgen.exception import SocketError
from mmgen.cfg import Config
from mmgen.rpc import rpc_init
from mmgen.util import make_timestr

async def get_rpc(cfg):
	try:
		return await rpc_init(cfg, ignore_wallet=True)
	except SocketError:
		return False

async def main(coins):

	rpcs = {}
	cfgs = {}
	test_suite = os.getenv('MMGEN_TEST_SUITE')
	base_cfg = Config({'pager':True})

	for coin in coins:
		cfg_in = {
			'coin': coin,
			'test_suite': test_suite,
		}
		if coin == 'eth' and not test_suite:
			cfg_in.update({'daemon_id': 'geth'})
		cfgs[coin] = Config(cfg_in)
		rpcs[coin] = await get_rpc(cfgs[coin])

	def gen_output():
		fs = '{:4} {:7} {:6} {:<5} {:<8} {:30} {:13} {:23} {}'
		yield fs.format(
			'Coin',
			'Network',
			'Status',
			'Port',
			'Chain',
			'Latest Block',
			'Daemon',
			'Version',
			'Datadir')
		for coin, rpc in rpcs.items():
			info = ('Down', '-', '-', '-', '-', '-', '-') if rpc is False else (
				'Up',
				rpc.port,
				rpc.chain,
				f'{rpc.blockcount:<8} [{make_timestr(rpc.cur_date)}]',
				rpc.daemon.coind_name,
				rpc.daemon_version_str,
				rpc.daemon.datadir
			)
			yield fs.format(coin.upper(), cfgs[coin].network, *info)

	base_cfg._util.stdout_or_pager('\n'.join(gen_output()))

all_coins = ('btc', 'ltc', 'eth')

coins = sys.argv[1:]

if coins and all(coin in all_coins for coin in coins):
	asyncio.run(main(coins))
else:
	print(f'You must supply one or more of the following coins on the command line:\n  {all_coins}')
	sys.exit(1)



#!/usr/bin/env python3

import sys, os
pn = os.path.abspath(os.path.dirname(sys.argv[0]))
os.chdir(os.path.dirname(os.path.dirname(pn)))
sys.path[0] = os.curdir

from mmgen.cfg import Config, gc
from mmgen.util import msg

opts_data = {
	'text': {
		'desc':    '',
		'usage':   'crypto | seed',
		'options': """
-h, --help            Print this help message
-P, --passwd-file=f   a
-p, --hash-preset=p   b
-r, --usr-randchars=n c
-L, --label=l         d
-m, --keep-label      e
		"""
	}
}

cfg = Config(opts_data=opts_data, init_opts={'color': True})

def crypto():
	desc = 'test data'

	from mmgen.crypto import Crypto
	crypto = Crypto(cfg)

	pw = crypto.get_new_passphrase(data_desc=desc, hash_preset=gc.dfl_hash_preset, passwd_file=None)
	msg(f'==> got new passphrase: [{pw}]\n')

	pw = crypto.get_passphrase(data_desc=desc, passwd_file=None)
	msg(f'==> got passphrase: [{pw}]\n')

	hp = crypto.get_hash_preset_from_user(data_desc=desc)
	msg(f'==> got hash preset: [{hp}]')

	hp = crypto.get_hash_preset_from_user(data_desc=desc)
	msg(f'==> got hash preset: [{hp}]')

def seed():

	from mmgen.wallet import Wallet

	for n in range(1, 3):
		msg(f'------- NEW WALLET {n} -------\n')
		w1 = Wallet(cfg)
		msg(f'\n==> got pw, preset, lbl: [{w1.ssdata.passwd}][{w1.ssdata.hash_preset}][{w1.ssdata.label}]\n')

	for n in range(1, 3):
		msg(f'------- PASSCHG {n} -------\n')
		w2 = Wallet(cfg, ss=w1, passchg=True)
		msg(f'\n==> got pw, preset, lbl: [{w2.ssdata.passwd}][{w2.ssdata.hash_preset}][{w2.ssdata.label}]\n')

	msg('------- WALLET FROM FILE -------\n')
	w3 = Wallet(cfg, fn='test/ref/FE3C6545-D782B529[128,1].mmdat') # passphrase: 'reference password'
	msg(f'\n==> got pw, preset, lbl: [{w3.ssdata.passwd}][{w3.ssdata.hash_preset}][{w3.ssdata.label}]\n')

globals()[cfg._args[0]]()



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
xmrwallet.ops.txview: Monero wallet ops for the MMGen Suite
"""

from pathlib import Path

from ...util import die

from ..file.tx import MoneroMMGenTX

from . import OpBase

class OpTxview(OpBase):
	view_method = 'get_info'
	opts = ('watch_only', 'autosign')
	hdr = ''
	col_hdr = ''
	footer = ''
	do_umount = False

	async def main(self, *, cols=None):

		self.mount_removable_device()

		if self.cfg.autosign:
			files = [f for f in self.asi.xmr_tx_dir.iterdir()
						if f.name.endswith('.'+MoneroMMGenTX.Submitted.ext)]
		else:
			files = self.uargs.infile

		txs = sorted(
			(MoneroMMGenTX.View(self.cfg, Path(fn)) for fn in files),
				# old TX files have no ‘submit_time’ field:
				key = lambda x: getattr(x.data, 'submit_time', None) or x.data.create_time
		)

		if self.cfg.autosign:
			self.asi.do_umount()

		addr_w = None if self.cfg.full_address or cols is None else cols - self.fixed_cols_w

		self.cfg._util.stdout_or_pager(
			(self.hdr if len(files) > 1 else '')
			+ self.col_hdr
			+ '\n'.join(getattr(tx, self.view_method)(addr_w=addr_w) for tx in txs)
			+ self.footer
		)

class OpTxlist(OpTxview):
	view_method = 'get_info_oneline'
	add_nl = True
	footer = '\n'
	fixed_cols_w = MoneroMMGenTX.Base.oneline_fixed_cols_w
	min_addr_w = 10

	@property
	def hdr(self):
		return ('SUBMITTED ' if self.cfg.autosign else '') + 'MONERO TRANSACTIONS\n'

	@property
	def col_hdr(self):
		return MoneroMMGenTX.View.oneline_fs.format(
			a = 'Network',
			b = 'Seed ID',
			c = 'Submitted' if self.cfg.autosign else 'Date',
			d = 'TxID',
			e = 'Type',
			f = 'Src',
			g = 'Dest',
			h = '  Amount',
			j = 'Dest Address',
			x = '',
		) + '\n'

	async def main(self):
		if self.cfg.pager:
			cols = None
		else:
			from ...term import get_terminal_size
			cols = self.cfg.columns or get_terminal_size().width
			if cols < self.fixed_cols_w + self.min_addr_w:
				die(1, f'A terminal at least {self.fixed_cols_w + self.min_addr_w} columns wide is required '
						'to display this output (or use --columns or --pager)')
		await super().main(cols=cols)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>

"""
test/colortest.py: test color handling for the MMGen suite
"""

import os

try:
	from include import test_init
except ImportError:
	from test.include import test_init

from mmgen.color import *
from mmgen.util import msg, ymsg, gmsg
import mmgen.color as color_mod

def test_color():

	ymsg("Terminal display:") # init_color() not called yet, so no yellow here

	for desc, nc in (
			('pre-init', None),
			('auto', 'auto'),
			('8-color', 8),
			('256-color', 256),
			('disabled', 0)):
		if nc is not None:
			init_color(num_colors=nc)
		msg('{:9}: {}'.format(
			desc,
			' '.join(getattr(color_mod, c)(c) for c in sorted(color_mod._colors))))

	init_color()
	gmsg("\nParsed terminfo 'colors' values:")

	from mmgen.color import orange
	for t, c in (
			('xterm', 8),
			('rxvt', 8),
			('rxvt-88color', 88),
			('rxvt-256color', 256),
			('screen-256color', 256),
			('xterm-256color', 256)):
		ret = get_terminfo_colors(t)
		if ret is None:
			ymsg(f'Warning: unable to get info for terminal {t!r}')
			continue
		msg(f'{t}: {orange(str(ret))}')
		assert c == ret, f"'colors' value for terminal {t} ({ret}) does not match expected value of {c}"

	ret = get_terminfo_colors()
	msg(f'{os.getenv("TERM")} (this terminal): {orange(str(ret))}')

if __name__ == '__main__':
	test_color()



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
altcointest.py - Test constants for Bitcoin-derived altcoins
"""

import sys

try:
	from include import test_init
except ImportError:
	from test.include import test_init

from mmgen.cfg import gc, Config
from mmgen.util import msg
from mmgen.altcoin.params import CoinInfo

def test_equal(desc, a, b, *cdata):
	if type(a) is int:
		a = hex(a)
		b = hex(b)
	(network, coin, _, b_desc, verbose) = cdata
	if verbose:
		msg(f'  {desc:20}: {a!r}')
	if a != b:
		raise ValueError(
			f'{desc.capitalize()}s for {coin.upper()} {network} do not match:\n  CoinInfo: {a}\n  {b_desc}: {b}')

class TestCoinInfo(CoinInfo):

	# Sources (see CoinInfo) that are in agreement for these coins
	# No check for segwit, p2sh check skipped if source doesn't support it
	cross_checks = {
		'2GIVE':  ['wn'],
		'42':     ['vg', 'wn'],
		'611':    ['wn'],
		'AC':     ['lb', 'vg'],
		'ACOIN':  ['wn'],
		'ALF':    ['wn'],
		'ANC':    ['vg', 'wn'],
		'APEX':   ['wn'],
		'ARCO':   ['wn'],
		'ARG':    ['pc'],
		'AUR':    ['vg', 'wn'],
		'BCH':    ['wn'],
		'BLK':    ['lb', 'vg', 'wn'],
		'BQC':    ['vg', 'wn'],
		'BSTY':   ['wn'],
		'BTC':    ['lb', 'vg', 'wn'],
		'BTCD':   ['lb', 'vg', 'wn'],
		'BUCKS':  ['wn'],
		'CASH':   ['wn'],
		'CBX':    ['wn'],
		'CCN':    ['lb', 'vg', 'wn'],
		'CDN':    ['lb', 'vg', 'wn'],
		'CHC':    ['wn'],
		'CLAM':   ['lb', 'vg'],
		'CON':    ['vg', 'wn'],
		'CPC':    ['wn'],
		'DASH':   ['lb', 'pc', 'vg', 'wn'],
		'DCR':    ['pc'],
		'DFC':    ['pc'],
		'DGB':    ['lb', 'vg'],
		'DGC':    ['lb', 'vg', 'wn'],
		'DOGE':   ['lb', 'pc', 'vg', 'wn'],
		'DOGED':  ['lb', 'vg', 'wn'],
		'DOPE':   ['lb', 'vg'],
		'DVC':    ['vg', 'wn'],
		'EFL':    ['lb', 'vg', 'wn'],
		'EMC':    ['vg'],
		'EMD':    ['wn'],
		'ESP':    ['wn'],
		'FAI':    ['pc'],
		'FC2':    ['wn'],
		'FIBRE':  ['wn'],
		'FJC':    ['wn'],
		'FLO':    ['wn'],
		'FLT':    ['wn'],
		'FST':    ['wn'],
		'FTC':    ['lb', 'pc', 'vg', 'wn'],
		'GCR':    ['lb', 'vg'],
		'GOOD':   ['wn'],
		'GRC':    ['vg', 'wn'],
		'GUN':    ['vg', 'wn'],
		'HAM':    ['vg', 'wn'],
		'HTML5':  ['wn'],
		'HYP':    ['wn'],
		'ICASH':  ['wn'],
		'INFX':   ['wn'],
		'IPC':    ['wn'],
		'JBS':    ['lb', 'pc', 'vg', 'wn'],
		'JUDGE':  ['wn'],
		'LANA':   ['wn'],
		'LAT':    ['wn'],
		'LDOGE':  ['wn'],
		'LMC':    ['wn'],
		'LTC':    ['lb', 'vg', 'wn'],
		'MARS':   ['wn'],
		'MEC':    ['pc', 'wn'],
		'MINT':   ['wn'],
		'MOBI':   ['wn'],
		'MONA':   ['lb', 'vg'],
		'MOON':   ['wn'],
		'MUE':    ['lb', 'vg'],
		'MXT':    ['wn'],
		'MYR':    ['pc'],
		'MYRIAD': ['vg', 'wn'],
		'MZC':    ['lb', 'pc', 'vg', 'wn'],
		'NEOS':   ['lb', 'vg'],
		'NEVA':   ['wn'],
		'NKA':    ['wn'],
		'NLG':    ['vg', 'wn'],
		'NMC':    ['lb', 'vg'],
		'NVC':    ['lb', 'vg', 'wn'],
		'OK':     ['lb', 'vg'],
		'OMC':    ['vg', 'wn'],
		'ONION':  ['vg', 'wn'],
		'PART':   ['wn'],
		'PINK':   ['vg', 'wn'],
		'PIVX':   ['wn'],
		'PKB':    ['lb', 'vg', 'wn'],
		'PND':    ['lb', 'vg', 'wn'],
		'POT':    ['lb', 'vg', 'wn'],
		'PPC':    ['lb', 'vg', 'wn'],
		'PTC':    ['vg', 'wn'],
		'PXC':    ['wn'],
		'QRK':    ['wn'],
		'RAIN':   ['wn'],
		'RBT':    ['wn'],
		'RBY':    ['lb', 'vg'],
		'RDD':    ['vg', 'wn'],
		'RIC':    ['pc', 'vg', 'wn'],
		'SDC':    ['lb', 'vg'],
		'SIB':    ['wn'],
		'SMLY':   ['wn'],
		'SONG':   ['wn'],
		'SPR':    ['vg', 'wn'],
		'START':  ['lb', 'vg'],
		'SYS':    ['wn'],
		'TAJ':    ['wn'],
		'TIT':    ['wn'],
		'TPC':    ['lb', 'vg'],
		'TRC':    ['wn'],
		'TTC':    ['wn'],
		'TX':     ['wn'],
		'UNO':    ['pc', 'vg', 'wn'],
		'VIA':    ['lb', 'pc', 'vg', 'wn'],
		'VPN':    ['lb', 'vg'],
		'VTC':    ['lb', 'vg', 'wn'],
		'WDC':    ['vg', 'wn'],
		'WISC':   ['wn'],
		'WKC':    ['vg', 'wn'],
		'WSX':    ['wn'],
		'XCN':    ['wn'],
		'XGB':    ['wn'],
		'XPM':    ['lb', 'vg', 'wn'],
		'XST':    ['wn'],
		'XVC':    ['wn'],
		'ZET':    ['wn'],
		'ZOOM':   ['lb', 'vg'],
		'ZRC':    ['lb', 'vg']
	}

	@classmethod
	def verify_leading_symbols(cls, quiet=False, verbose=False):

		for network in ('mainnet', 'testnet'):
			for coin in [e.symbol for e in cls.coin_constants[network]]:
				e = cls.get_entry(coin, network)
				cdata = (network, coin, e, 'Computed value', verbose)

				if not quiet:
					msg(f'{coin} {network}')

				vn_info = e.p2pkh_info
				ret = cls.find_addr_leading_symbol(vn_info[0])
				test_equal('P2PKH leading symbol', vn_info[1], ret, *cdata)

				vn_info = e.p2sh_info
				if vn_info:
					ret = cls.find_addr_leading_symbol(vn_info[0])
					test_equal('P2SH leading symbol', vn_info[1], ret, *cdata)

	@classmethod
	def verify_core_coin_data(cls, cfg, quiet=False, verbose=False):
		from mmgen.protocol import CoinProtocol, init_proto

		for network in ('mainnet', 'testnet'):
			for coin in gc.core_coins:
				e = cls.get_entry(coin, network)
				if e:
					proto = init_proto(cfg, coin, network=network)
					cdata = (network, coin, e, type(proto).__name__, verbose)
					if not quiet:
						msg(f'Verifying {coin.upper()} {network}')

					if coin != 'bch': # TODO
						test_equal('coin name', e.name, proto.name, *cdata)

					if e.trust_level != -1:
						test_equal('Trust level', e.trust_level, CoinProtocol.coins[coin].trust_level, *cdata)

					test_equal(
						'WIF version number',
						e.wif_ver_num,
						int.from_bytes(proto.wif_ver_bytes['std'], 'big'),
						*cdata)

					test_equal(
						'P2PKH version number',
						e.p2pkh_info[0],
						int.from_bytes(proto.addr_fmt_to_ver_bytes['p2pkh'], 'big'),
						*cdata)

					test_equal(
						'P2SH version number',
						e.p2sh_info[0],
						int.from_bytes(proto.addr_fmt_to_ver_bytes['p2sh'], 'big'),
						*cdata)

	# Data is one of the coin_constants lists above.  Normalize ints to hex of correct width, add
	# missing leading letters, set trust level from external_tests.
	# Insert a coin entry from outside source, set version info leading letters to '?' and trust level
	# to 0, then run TestCoinInfo.fix_table(data).  'has_segwit' field is updated manually for now.
	@classmethod
	def fix_table(cls, data):
		import re

		def myhex(n):
			return '0x{:0{}x}'.format(n, 2 if n < 256 else 4)

		def fix_ver_info(e, k):
			e[k] = list(e[k])
			e[k][0] = myhex(e[k][0])
			s1 = cls.find_addr_leading_symbol(int(e[k][0][2:], 16))
			m = f'Fixing leading address letter for coin {e["symbol"]} ({e[k][1]!r} --> {s1})'
			if e[k][1] != '?':
				assert s1 == e[k][1], f'First letters do not match! {m}'
			else:
				msg(m)
				e[k][1] = s1
			e[k] = tuple(e[k])

		old_sym = None
		for sym in sorted([e.symbol for e in data]):
			if sym == old_sym:
				msg(f'{sym!r}: duplicate coin symbol in data!')
				sys.exit(2)
			old_sym = sym

		tt = cls.create_trust_table()

		name_w = max(len(e.name) for e in data)
		fs = '\t({:%s} {:10} {:7} {:17} {:17} {:6} {}),' % (name_w+3)
		for e in data:
			e = e._asdict()
			e['wif_ver_num'] = myhex(e['wif_ver_num'])
			sym, trust = e['symbol'], e['trust_level']

			fix_ver_info(e, 'p2pkh_info')
			if isinstance(e['p2sh_info'], tuple):
				fix_ver_info(e, 'p2sh_info')

			for k in e.keys():
				e[k] = repr(e[k])
				e[k] = re.sub(r"'0x(..)'", r'0x\1', e[k])
				e[k] = re.sub(r"'0x(....)'", r'0x\1', e[k])
				e[k] = re.sub(r' ', r'', e[k]) + ('', ',')[k != 'trust_level']

			if trust != -1:
				if sym in tt:
					src = tt[sym]
					if src != trust:
						msg(f'Updating trust for coin {sym!r}: {trust} -> {src}')
						e['trust_level'] = src
				else:
					if trust != 0:
						msg(f'Downgrading trust for coin {sym!r}: {trust} -> 0')
						e['trust_level'] = 0

				if sym in cls.cross_checks:
					if int(e['trust_level']) == 0 and len(cls.cross_checks[sym]) > 1:
						msg(f'Upgrading trust for coin {sym!r}: {e["trust_level"]} -> 1')
						e['trust_level'] = 1

			print(fs.format(*e.values()))
		msg(f'Processed {len(data)} entries')

	@classmethod
	def find_addr_leading_symbol(cls, ver_num, verbose=False):

		if ver_num == 0:
			return '1'

		def phash2addr(ver_num, pk_hash):
			from mmgen.proto.btc.common import b58chk_encode
			bl = ver_num.bit_length()
			ver_bytes = int.to_bytes(ver_num, bl//8 + bool(bl%8), 'big')
			return b58chk_encode(ver_bytes + pk_hash)

		low = phash2addr(ver_num, b'\x00'*20)
		high = phash2addr(ver_num, b'\xff'*20)

		if verbose:
			print('low address:  ' + low)
			print('high address: ' + high)

		l1, h1 = low[0], high[0]
		return (l1, h1) if l1 != h1 else l1

	@classmethod
	def print_symbols(cls, include_names=False, reverse=False):
		for e in cls.coin_constants['mainnet']:
			if reverse:
				print(f'{e.symbol:6} {e.name}')
			else:
				name_w = max(len(e.name) for e in cls.coin_constants['mainnet'])
				print((f'{e.name:{name_w}} ' if include_names else '') + e.symbol)

	@classmethod
	def create_trust_table(cls):
		tt = {}
		mn = cls.external_tests['mainnet']
		for ext_prog in mn:
			assert len(set(mn[ext_prog])) == len(mn[ext_prog]), f'Duplicate entry in {ext_prog!r}!'
			for coin in mn[ext_prog]:
				if coin in tt:
					tt[coin] += 1
				else:
					tt[coin] = 1
		for k in cls.trust_override:
			tt[k] = cls.trust_override[k]
		return tt

	trust_override = {'BTC':3, 'BCH':3, 'LTC':3, 'DASH':1, 'EMC':2}

	@classmethod
	def get_test_support(cls, coin, addr_type, network, toolname=None, verbose=False):
		"""
		If requested tool supports coin/addr_type/network triplet, return tool name.
		If 'tool' is None, return tool that supports coin/addr_type/network triplet.
		Return None on failure.
		"""
		all_tools = [toolname] if toolname else list(cls.external_tests[network].keys())
		coin = coin.upper()

		for tool in all_tools:
			if coin in cls.external_tests[network][tool]:
				break
		else:
			if verbose:
				m1 = 'Requested tool {t!r} does not support coin {c} on network {n}'
				m2 = 'No test tool found for coin {c} on network {n}'
				msg((m1 if toolname else m2).format(t=tool, c=coin, n=network))
			return None

		if addr_type == 'zcash_z':
			if toolname in (None, 'zcash-mini'):
				return 'zcash-mini'
			else:
				if verbose:
					msg(f"Address type {addr_type!r} supported only by tool 'zcash-mini'")
				return None

		try:
			bl = cls.external_tests_blacklist[addr_type][tool]
		except:
			pass
		else:
			if bl is True or coin in bl:
				if verbose:
					msg(f'Tool {tool!r} blacklisted for coin {coin}, addr_type {addr_type!r}')
				return None

		if toolname: # skip whitelists
			return tool

		if addr_type in ('segwit', 'bech32'):
			st = cls.external_tests_segwit_whitelist
			if addr_type in st and coin in st[addr_type]:
				return tool
			else:
				if verbose:
					m1 = 'Requested tool {t!r} does not support coin {c}, addr_type {a!r}, on network {n}'
					m2 = 'No test tool found supporting coin {c}, addr_type {a!r}, on network {n}'
					msg((m1 if toolname else m2).format(t=tool, c=coin, n=network, a=addr_type))
				return None

		return tool

	external_tests = {
		'mainnet': {
			# List in order of preference.
			# If 'tool' is not specified, the first tool supporting the coin will be selected.
			'pycoin': (
				'DASH', # only compressed
				'BCH',
				'BTC', 'LTC', 'VIA', 'FTC', 'DOGE', 'MEC',
				'JBS', 'MZC', 'RIC', 'DFC', 'FAI', 'ARG', 'ZEC', 'DCR'),
			'keyconv': ( # broken: PIVX
				'BCH', '42', 'AC', 'AIB', 'ANC', 'ARS', 'ATMOS', 'AUR', 'BLK', 'BQC', 'BTC', 'TEST',
				'BTCD', 'CCC', 'CCN', 'CDN', 'CLAM', 'CNC', 'CNOTE', 'CON', 'CRW', 'DEEPONION', 'DGB',
				'DGC', 'DMD', 'DOGED', 'DOGE', 'DOPE', 'DVC', 'EFL', 'EMC', 'EXCL', 'FAIR', 'FLOZ', 'FTC',
				'GAME', 'GAP', 'GCR', 'GRC', 'GRS', 'GUN', 'HAM', 'HODL', 'IXC', 'JBS', 'LBRY', 'LEAF',
				'LTC', 'MMC', 'MONA', 'MUE', 'MYRIAD', 'MZC', 'NEOS', 'NLG', 'NMC', 'NVC', 'NYAN', 'OK',
				'OMC', 'PIGGY', 'PINK', 'PKB', 'PND', 'POT', 'PPC', 'PTC', 'PTS', 'QTUM', 'RBY', 'RDD',
				'RIC', 'SCA', 'SDC', 'SKC', 'SPR', 'START', 'SXC', 'TPC', 'UIS', 'UNO', 'VIA', 'VPN',
				'VTC', 'WDC', 'WKC', 'WUBS', 'XC', 'XPM', 'YAC', 'ZOOM', 'ZRC'),
			'eth-keys': ('ETH', 'ETC'),
			'zcash-mini': ('ZEC',),
			'monero-python': ('XMR',),
		},
		'testnet': {
			'pycoin': {
				'DASH':'tDASH', # only compressed
				'BCH':'XTN',
				'BTC':'XTN', 'LTC':'XLT', 'VIA':'TVI', 'FTC':'FTX', 'DOGE':'XDT', 'DCR':'DCRT'
				},
			'eth-keys': {},
			'keyconv': {}
		}
	}
	external_tests_segwit_whitelist = {
		# Whitelists apply to the *first* tool in cls.external_tests supporting the given coin/addr_type.
		# They're ignored if specific tool is requested.
		'segwit': ('BTC',), # LTC Segwit broken on pycoin: uses old fmt
		'bech32': ('BTC', 'LTC'),
		'compressed': (
			'BTC', 'LTC', 'VIA', 'FTC', 'DOGE', 'DASH', 'MEC', 'MYR', 'UNO',
			'JBS', 'MZC', 'RIC', 'DFC', 'FAI', 'ARG', 'ZEC', 'DCR', 'ZEC'
		),
	}
	external_tests_blacklist = {
		# Unconditionally block testing of the given coin/addr_type with given tool, or all coins if True
		'legacy': {},
		'segwit': {'keyconv': True},
		'bech32': {'keyconv': True},
	}

if __name__ == '__main__':

	opts_data = {
		'text': {
			'desc': 'Check altcoin data',
			'usage':'[opts]',
			'options': '-q, --quiet    Be quieter\n-v, --verbose  Be more verbose'
		}
	}

	cfg = Config(opts_data=opts_data, need_amt=False)

	msg('Checking CoinInfo WIF/P2PKH/P2SH version numbers and trust levels against protocol.py')
	TestCoinInfo.verify_core_coin_data(cfg, cfg.quiet, cfg.verbose)

	msg('Checking CoinInfo address leading symbols')
	TestCoinInfo.verify_leading_symbols(cfg.quiet, cfg.verbose)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
mmgen-xmrwallet: Perform various Monero wallet and transacting operations for
                 addresses in an MMGen XMR key-address file
"""
import asyncio

from .cfg import gc, Config
from .util import die, fmt_dict
from . import xmrwallet

opts_data = {
	'sets': [
		('autosign', True, 'watch_only', True),
		('autosign_mountpoint', bool, 'autosign', True),
		('autosign_mountpoint', bool, 'watch_only', True),
	],
	'text': {
		'desc': """Perform various Monero wallet and transacting operations for
                   addresses in an MMGen XMR key-address file""",
		'usage2': [
			'[opts] create | sync | list | view | listview | dump | restore [xmr_keyaddrfile] [wallets]',
			'[opts] label    [xmr_keyaddrfile] LABEL_SPEC',
			'[opts] new      [xmr_keyaddrfile] NEW_ADDRESS_SPEC',
			'[opts] transfer [xmr_keyaddrfile] TRANSFER_SPEC',
			'[opts] sweep | sweep_all [xmr_keyaddrfile] SWEEP_SPEC',
			'[opts] submit   [TX_file]',
			'[opts] relay    <TX_file>',
			'[opts] resubmit | abort (for use with --autosign only)',
			'[opts] txview | txlist [TX_file] ...',
			'[opts] export-outputs | export-outputs-sign | import-key-images [wallets]',
		],
		'options': """
-h, --help                       Print this help message
--, --longhelp                   Print help message for long (global) options
-a, --autosign                   Use appropriate outdir and other params for
                                 autosigning operations (implies --watch-only).
                                 When this option is in effect, filename argu-
                                 ments must be omitted, as files are located
                                 automatically.
-f, --priority=N                 Specify an integer priority ‘N’ for inclusion
                                 of a transaction in the blockchain (higher
                                 number means higher fee).  Valid parameters:
                                 {tp}.  If option
                                 is omitted, the default priority will be used
-F, --full-address               Print addresses in full instead of truncating
-m, --autosign-mountpoint=P      Specify the autosign mountpoint (defaults to
                                 ‘/mnt/mmgen_autosign’, implies --autosign)
-b, --rescan-blockchain          Rescan the blockchain if wallet fails to sync
-d, --outdir=D                   Save transaction files to directory 'D'
                                 instead of the working directory
-D, --daemon=H:P                 Connect to the monerod at {D}
-e, --skip-empty-accounts        Skip display of empty accounts in wallets
                                 where applicable
-E, --skip-empty-addresses       Skip display of used empty addresses in
                                 wallets where applicable
-k, --use-internal-keccak-module Force use of the internal keccak module
-p, --hash-preset=P              Use scrypt hash preset 'P' for password
                                 hashing (default: '{gc.dfl_hash_preset}')
-P, --rescan-spent               Perform a rescan of spent outputs.  Used only
                                 with the ‘export-outputs-sign’ operation
-R, --tx-relay-daemon=H:P[:H:P]  Relay transactions via a monerod specified by
                                 {R}
-r, --restore-height=H           Scan from height 'H' when creating wallets.
                                 Use special value ‘current’ to create empty
                                 wallet at current blockchain height.
-R, --no-relay                   Save transaction to file instead of relaying
-s, --no-start-wallet-daemon     Don’t start the wallet daemon at startup
-S, --no-stop-wallet-daemon      Don’t stop the wallet daemon at exit
-W, --watch-only                 Create or operate on watch-only wallets
-w, --wallet-dir=D               Output or operate on wallets in directory 'D'
                                 instead of the working directory
-U, --wallet-rpc-user=user       Wallet RPC username (currently: {cfg.monero_wallet_rpc_user!r})
-P, --wallet-rpc-password=pass   Wallet RPC password (currently: [scrubbed])
""",
	'notes': """

{xmrwallet_help}
"""
	},
	'code': {
		'options': lambda cfg, s: s.format(
			D   = xmrwallet.uarg_info['daemon'].annot,
			R   = xmrwallet.uarg_info['tx_relay_daemon'].annot,
			cfg = cfg,
			gc  = gc,
			tp  = fmt_dict(xmrwallet.tx_priorities, fmt='equal_compact')
		),
		'notes': lambda help_mod, s: s.format(
			xmrwallet_help = help_mod('xmrwallet')
		)
	}
}

cfg = Config(opts_data=opts_data, init_opts={'coin':'xmr'})

cmd_args = cfg._args

if cmd_args and cfg.autosign and (
		cmd_args[0] in (
			xmrwallet.kafile_arg_ops
			+ ('export-outputs', 'export-outputs-sign', 'import-key-images', 'txview', 'txlist')
		)
		or len(cmd_args) == 1 and cmd_args[0] in ('submit', 'resubmit', 'abort')
	):
	cmd_args.insert(1, None)

if len(cmd_args) < 2:
	cfg._usage()

op     = cmd_args.pop(0)
infile = cmd_args.pop(0)
wallets = spec = None

if op in ('relay', 'submit', 'resubmit', 'abort'):
	if len(cmd_args) != 0:
		cfg._usage()
elif op in ('txview', 'txlist'):
	infile = [infile] + cmd_args
elif op in ('create', 'sync', 'list', 'view', 'listview', 'dump', 'restore'): # kafile_arg_ops
	if len(cmd_args) > 1:
		cfg._usage()
	wallets = cmd_args.pop(0) if cmd_args else None
elif op in ('new', 'transfer', 'sweep', 'sweep_all', 'label'):
	if len(cmd_args) != 1:
		cfg._usage()
	spec = cmd_args[0]
elif op in ('export-outputs', 'export-outputs-sign', 'import-key-images'):
	if not cfg.autosign:
		die(1, f'--autosign must be used with command {op!r}')
	if len(cmd_args) > 1:
		cfg._usage()
	wallets = cmd_args.pop(0) if cmd_args else None
else:
	die(1, f'{op!r}: unrecognized operation')

m = xmrwallet.op(op, cfg, infile, wallets, spec=spec)

if asyncio.run(m.main()):
	m.post_main_success()
else:
	m.post_main_failure()



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
tw.json: export and import tracking wallet to JSON format
"""

import sys, os, json
from collections import namedtuple

from ..util import msg, ymsg, fmt, suf, die, make_timestamp, make_chksum_8
from ..base_obj import AsyncInit
from ..objmethods import MMGenObject
from ..rpc.util import json_encoder
from .ctl import TwCtl

class TwJSON:

	class Base(MMGenObject):

		can_prune = False
		pruned = None
		fn_pfx = 'mmgen-tracking-wallet-dump'

		def __new__(cls, cfg, proto, *args, **kwargs):
			return MMGenObject.__new__(
				proto.base_proto_subclass(TwJSON, 'tw.json', sub_clsname=cls.__name__))

		def __init__(self, cfg, proto):
			self.cfg = cfg
			self.proto = proto
			self.coin = proto.coin_id.lower()
			self.network = proto.network
			self.keys = ['mmgen_id', 'address', 'amount', 'comment']
			self.entry_tuple = namedtuple('tw_entry', self.keys)

		@property
		def dump_fn(self):

			def get_fn(prune_id):
				return '{a}{b}-{c}-{d}.json'.format(
					a = self.fn_pfx,
					b = f'-pruned[{prune_id}]' if prune_id else '',
					c = self.coin,
					d = self.network)

			if self.pruned:
				from ..addrlist import AddrIdxList
				prune_id = AddrIdxList(idx_list=self.pruned).id_str
				fn = get_fn(prune_id)
				mf = 255 if sys.platform == 'win32' else os.statvfs(self.cfg.outdir or os.curdir).f_namemax
				if len(fn) > mf:
					fn = get_fn(f'idhash={make_chksum_8(prune_id.encode()).lower()}')
			else:
				fn = get_fn(None)

			return fn

		def json_dump(self, data, *, pretty=False):
			return json.dumps(
				data,
				cls        = json_encoder,
				sort_keys  = True,
				separators = None if pretty else (',', ':'),
				indent     = 4 if pretty else None) + ('\n' if pretty else '')

		def make_chksum(self, data):
			return make_chksum_8(self.json_dump(data).encode()).lower()

		@property
		def mappings_chksum(self):
			return self.make_chksum(self.mappings_json)

		@property
		def entry_tuple_in(self):
			return namedtuple('entry_tuple_in', self.keys)

	class Import(Base, metaclass=AsyncInit):

		blockchain_rescan_warning = None

		async def __init__(
				self,
				cfg,
				proto,
				filename,
				*,
				ignore_checksum = False,
				batch           = False):

			super().__init__(cfg, proto)

			self.twctl = await TwCtl(cfg, proto, mode='i', rpc_ignore_wallet=True)

			def check_network(data):
				coin, network = data['network'].split('_')
				if coin != self.coin:
					die(2, f'Coin in wallet dump is {coin.upper()}, but configured coin is {self.coin.upper()}')
				if network != self.network:
					die(2, f'Network in wallet dump is {network}, but configured network is {self.network}')

			def check_chksum(d):
				chksum = self.make_chksum(d['data'])
				if chksum != d['checksum']:
					if ignore_checksum:
						ymsg(f'Warning: ignoring incorrect checksum {chksum}')
					else:
						die(3, f'File checksum incorrect! ({chksum} != {d["checksum"]})')

			def verify_data(d):
				check_network(d['data'])
				check_chksum(d)
				self.cfg._util.compare_or_die(
					val1  = self.mappings_chksum,
					val2  = d['data']['mappings_checksum'],
					desc1 = 'computed mappings checksum',
					desc2 = 'saved checksum')

			if not await self.check_and_create_wallet():
				return

			from ..fileutil import get_data_from_file
			self.data = json.loads(get_data_from_file(self.cfg, filename, quiet=True))
			self.keys = self.data['data']['entries_keys']
			self.entries = await self.get_entries()

			verify_data(self.data)

			addrs = await self.do_import(batch)

			await self.twctl.rescan_addresses(addrs)

			if self.blockchain_rescan_warning:
				ymsg('\n' + fmt(self.blockchain_rescan_warning.strip(), indent='  '))

		async def check_and_create_wallet(self):

			if await self.tracking_wallet_exists:
				die(3,
					f'Existing {self.twctl.rpc.daemon.desc} wallet detected!\n' +
					'It must be moved, or backed up and securely deleted, before running this command')

			msg('\n'+fmt(self.info_msg.strip(), indent='  '))

			from ..ui import keypress_confirm
			if not keypress_confirm(self.cfg, 'Continue?'):
				msg('Exiting at user request')
				return False

			if not await self.create_tracking_wallet():
				die(3, 'Wallet could not be created')

			return True

	class Export(Base, metaclass=AsyncInit):

		async def __init__(
				self,
				cfg,
				proto,
				*,
				include_amts    = True,
				pretty          = False,
				prune           = False,
				warn_used       = False,
				force_overwrite = False):

			if prune and not self.can_prune:
				die(1, f'Pruning not supported for {proto.name} protocol')

			self.prune = prune
			self.warn_used = warn_used

			super().__init__(cfg, proto)

			if not include_amts:
				self.keys.remove('amount')

			self.twctl = await TwCtl(cfg, proto)

			self.entries = await self.get_entries()

			if self.prune:
				msg('Pruned {} address{}'.format(len(self.pruned), suf(self.pruned, 'es')))

			msg('Exporting {} address{}'.format(self.num_entries, suf(self.num_entries, 'es')))

			data = {
				'id': 'mmgen_tracking_wallet',
				'version': 1,
				'network': f'{self.coin}_{self.network}',
				'blockheight': self.twctl.rpc.blockcount,
				'time': make_timestamp(),
				'mappings_checksum': self.mappings_chksum,
				'entries_keys': self.keys,
				'entries': await self.entries_out,
				'num_entries': self.num_entries,
			}
			if include_amts:
				data['value'] = await self.total

			from ..fileutil import write_data_to_file
			write_data_to_file(
				cfg     = self.cfg,
				outfile = self.dump_fn,
				data    = self.json_dump(
					{
						'checksum': self.make_chksum(data),
						'data': data
					},
					pretty = pretty),
				desc    = 'tracking wallet JSON data',
				ask_overwrite = not force_overwrite)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
scripts/create-token.py: Automated ERC20 token creation for the MMGen suite
"""

import sys, json, re
from subprocess import run, PIPE
from collections import namedtuple

import script_init
from mmgen.main import launch
from mmgen.cfg import Config
from mmgen.util import Msg, msg, rmsg, ymsg, die

ti = namedtuple('token_param_info', ['default', 'conversion', 'test'])
class TokenData:
	fields = ('decimals', 'supply', 'name', 'symbol', 'owner_addr')
	decimals   = ti('18', int, lambda s: s.isascii() and s.isdigit() and 0 < int(s) <= 36)
	name       = ti(None, str, lambda s: s.isascii() and s.isprintable() and len(s) < 256)
	supply     = ti(None, int, lambda s: s.isascii() and s.isdigit() and 0 < int(s) < 2**256)
	symbol     = ti(None, str, lambda s: s.isascii() and s.isalnum() and len(s) <= 20)
	owner_addr = ti(None, str, lambda s: s.isascii() and s.isalnum() and len(s) == 40) # checked separately

token_data = TokenData()

req_solc_ver_pat = '^0.8.25'

opts_data = {
	'text': {
		'desc': 'Create an ERC20 token contract',
		'usage':'[opts] <owner address>',
		'options': f"""
-h, --help        Print this help message
-o, --outdir=D    Specify output directory for *.bin files
-d, --decimals=D  Number of decimals for the token (default: {token_data.decimals.default})
-n, --name=N      Token name (REQUIRED)
-p, --preprocess  Print the preprocessed code to stdout
-t, --supply=T    Total supply of the token (REQUIRED)
-s, --symbol=S    Token symbol (REQUIRED)
-S, --stdout      Output JSON data to stdout instead of files
-v, --verbose     Produce more verbose output
-c, --check-solc-version Check the installed solc version
""",
	'notes': """
The owner address must be in checksummed format.

Use ‘mmgen-tool eth_checksummed_addr’ to create it if necessary.
"""
	}
}

import os
with open(os.path.join(os.path.dirname(__file__), 'ERC20.sol.in')) as fh:
	solidity_code_template = fh.read()

def create_src(cfg, template, token_data):

	def gen():
		for k in token_data.fields:
			field = getattr(token_data, k)
			if k == 'owner_addr':
				owner_addr = cfg._args[0]
				from mmgen.addr import is_coin_addr
				if not is_coin_addr(cfg._proto, owner_addr.lower()):
					die(1, f'{owner_addr}: not a valid {cfg._proto.coin} coin address')
				val = '0x' + owner_addr
			else:
				val = (
					getattr(cfg, k)
					or getattr(field, 'default', None)
					or die(1, f'The --{k} option must be specified')
				)
				if not field.test(val):
					die(1, f'{val!r}: invalid parameter for option --{k}')

			yield (k, field.conversion(val))

	from string import Template
	return Template(template).substitute(**(dict(gen()) | {'solc_ver_pat': req_solc_ver_pat}))

def check_solc_version():
	"""
	The output is used by other programs, so write to stdout only
	"""
	try:
		cp = run(['solc', '--version'], check=True, stdout=PIPE)
	except:
		msg('solc missing or could not be executed') # this must go to stderr
		return False

	if cp.returncode != 0:
		Msg('solc exited with error')
		return False

	line = cp.stdout.decode().splitlines()[1]
	version_str = re.sub(r'Version:\s*', '', line)
	m = re.match(r'(\d+)\.(\d+)\.(\d+)', version_str)

	if not m:
		Msg(f'Unrecognized solc version string: {version_str}')
		return False

	from semantic_version import Version, NpmSpec
	version = Version('{}.{}.{}'.format(*m.groups()))

	if version in NpmSpec(req_solc_ver_pat):
		Msg(str(version))
		return True
	else:
		Msg(f'solc version ({version_str}) does not match requirement ({req_solc_ver_pat})')
		return False

def compile_code(cfg, code):
	cmd = ['solc', '--optimize', '--bin', '--overwrite', '--evm-version=constantinople']
	if not cfg.stdout:
		cmd += ['--output-dir', cfg.outdir or '.']
	cmd += ['-']
	msg(f"Executing: {' '.join(cmd)}")
	cp = run(cmd, input=code.encode(), stdout=PIPE, stderr=PIPE)
	out = cp.stdout.decode().replace('\r', '')
	err = cp.stderr.decode().replace('\r', '').strip()
	if cp.returncode != 0:
		rmsg('Solidity compiler produced the following error:')
		msg(err)
		die(4, f'Solidity compiler exited with error (return val: {cp.returncode})')
	if err:
		ymsg('Solidity compiler produced the following warning:')
		msg(err)
	if cfg.stdout:
		o = out.split('\n')
		return {k:o[i+2] for k in ('SafeMath', 'Owned', 'Token') for i in range(len(o)) if k in o[i]}
	else:
		cfg._util.vmsg(out)

def main():
	cfg = Config(opts_data=opts_data)

	if cfg.check_solc_version:
		sys.exit(0 if check_solc_version() else 1)

	if not cfg._proto.coin in ('ETH', 'ETC'):
		die(1, '--coin option must be ETH or ETC')

	if not len(cfg._args) == 1:
		cfg._usage()

	code = create_src(cfg, solidity_code_template, token_data)

	if cfg.preprocess:
		Msg(code)
		sys.exit(0)

	out = compile_code(cfg, code)

	if cfg.stdout:
		print(json.dumps(out))

	msg('Contract successfully compiled')

if __name__ == '__main__':

	launch(func=main)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
xmrwallet.ops.import: Monero wallet ops for the MMGen Suite
"""

from ...util import msg, bmsg, die, suf

from ..file.outputs import MoneroWalletOutputsFile
from ..rpc import MoneroWalletRPC

from .wallet import OpWallet

class OpImportOutputs(OpWallet):
	action = 'importing wallet outputs into'
	start_daemon = False

	async def main(self, fn, wallet_idx, *, restart_daemon=True):
		if restart_daemon:
			await self.restart_wallet_daemon()
		h = MoneroWalletRPC(self, self.addr_data[0])
		self.head_msg(wallet_idx, fn)
		if restart_daemon:
			h.open_wallet(refresh=False)
		m = MoneroWalletOutputsFile.Unsigned(
			parent = self,
			fn     = fn)
		res = self.c.call(
			'import_outputs',
			outputs_data_hex = m.data.outputs_data_hex)
		idata = res['num_imported']
		bmsg(f'\n  {idata} output{suf(idata)} imported')
		if m.data.sign:
			data = m.data._asdict()
			data.update(self.c.call('export_key_images', all=True))
			m = MoneroWalletOutputsFile.SignedNew(
				parent    = self,
				wallet_fn = m.get_wallet_fn(fn),
				data      = data)
			idata = m.data.signed_key_images or []
			bmsg(f'  {len(idata)} key image{suf(idata)} signed')
		else:
			m.data = m.data._replace(imported=True)
		return m

class OpImportKeyImages(OpWallet):
	action = 'importing key images into'
	stem = 'process'
	trust_monerod = True

	def post_main_failure(self):
		rw_msg = ' for requested wallets' if self.uargs.wallets else ''
		die(2, f'No signed key image files found{rw_msg}!')

	async def process_wallet(self, d, fn, last):
		keyimage_fn = MoneroWalletOutputsFile.Signed.find_fn_from_wallet_fn(self.cfg, fn, ret_on_no_match=True)
		if not keyimage_fn:
			msg(f'No signed key image file found for wallet #{d.idx}')
			return False
		h = MoneroWalletRPC(self, d)
		h.open_wallet()
		self.head_msg(d.idx, h.fn)
		m = MoneroWalletOutputsFile.Signed(parent=self, fn=keyimage_fn)
		data = m.data.signed_key_images or []
		bmsg(f'\n  {len(data)} signed key image{suf(data)} to import')
		if data:
			res = self.c.call('import_key_images', signed_key_images=data)
			bmsg(f'  Success: {res}')
		return True



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
tx.bump: transaction bump class
"""

from .new_swap import NewSwap
from .completed import Completed
from ..util import msg, ymsg, is_int, die
from ..color import orange, pink

class Bump(Completed, NewSwap):
	desc = 'fee-bumped transaction'
	ext  = 'rawtx'
	bump_output_idx = None
	is_bump = True

	def __init__(self, *, check_sent, new_outputs, **kwargs):

		super().__init__(**kwargs)

		self.new_outputs = new_outputs
		self.orig_rel_fee = self.get_orig_rel_fee()

		if new_outputs:
			if self.is_swap:
				self.is_swap = False
				for attr in self.swap_attrs:
					setattr(self, attr, None)
			self.outputs = self.OutputList(self)
			self.cfg = kwargs['cfg'] # must use current cfg opts, not those from orig_tx
		elif self.is_swap and self.is_token:
			die(1,
				orange('Fee-bumping of token swap transactions currently not supported.\n') +
				orange('To bump the transaction, supply new outputs on the command line'))

		if not self.is_replaceable():
			die(1, f'Transaction {self.txid} is not replaceable')

		# If sending, require original tx to be sent
		if check_sent and not self.coin_txid:
			die(1, f'Transaction {self.txid!r} was not broadcast to the network')

		self.coin_txid = ''
		self.sent_timestamp = None

	async def get_inputs(self, outputs_sum):
		return True

	def check_bumped_fee_ok(self, abs_fee):
		from ..amt import RelFeeAmt
		orig = RelFeeAmt(self.orig_rel_fee)
		new = RelFeeAmt(self.fee_abs2rel(abs_fee))
		if new <= orig:
			fs = 'New fee ({b!s} {d}) <= original fee ({a!s} {d})\nPlease choose a higher fee'
			ymsg(fs.format(a=orig, b=new, d=self.rel_fee_disp))
			return False
		return True

	async def create_feebump(self, silent):

		from ..rpc import rpc_init
		self.rpc = await rpc_init(self.cfg, self.proto)

		msg('Creating replacement transaction')

		self.check_sufficient_funds_for_bump()

		output_idx = self.choose_output()

		await self.set_gas(force=True)

		if not silent:
			msg('Minimum fee for new transaction: {} {} ({} {})'.format(
				self.min_fee.hl(),
				self.proto.coin,
				pink(self.fee_abs2rel(self.min_fee)),
				self.rel_fee_disp))

		if self.is_swap:
			self.recv_proto = self.check_swap_memo().proto
			self.swap_cfg = self.swap_proto_mod.SwapCfg(self.cfg)
			fee_hint = await self.update_vault_output(self.send_amt)
		else:
			fee_hint = None

		self.usr_fee = self.get_usr_fee_interactive(
			fee = self.cfg.fee or fee_hint,
			desc = 'User-selected' if self.cfg.fee else 'Recommended' if fee_hint else None)

		self.bump_fee(output_idx, self.usr_fee)

		assert self.fee <= self.proto.max_tx_fee

		if not self.cfg.yes:
			self.add_comment()   # edits an existing comment

		await self.create_serialized()

		self.add_timestamp()
		self.add_blockcount()

		self.cfg._util.qmsg('Fee successfully increased')

	def check_sufficient_funds_for_bump(self):
		if not [o.amt for o in self.outputs if o.amt >= self.min_fee]:
			die(1,
				'Transaction cannot be bumped.\n' +
				f'All outputs contain less than the minimum fee ({self.min_fee} {self.coin})')

	def choose_output(self):

		def check_sufficient_funds(o_amt):
			if o_amt < self.min_fee:
				msg(f'Minimum fee ({self.min_fee} {self.coin}) is greater than output amount ({o_amt} {self.coin})')
				return False
			return True

		if len(self.nondata_outputs) == 1:
			if check_sufficient_funds(self.nondata_outputs[0].amt):
				self.bump_output_idx = 0
				return 0
			else:
				die(1, 'Insufficient funds to bump transaction')

		init_reply = self.cfg.output_to_reduce
		chg_idx = self.chg_idx

		while True:
			if init_reply is None:
				from ..ui import line_input
				m = 'Choose an output to deduct the fee from (Hit ENTER for the change output): '
				reply = line_input(self.cfg, m) or 'c'
			else:
				reply, init_reply = init_reply, None
			if chg_idx is None and not is_int(reply):
				msg('Output must be an integer')
			elif chg_idx is not None and not is_int(reply) and reply != 'c':
				msg("Output must be an integer, or 'c' for the change output")
			else:
				idx = chg_idx if reply == 'c' else (int(reply) - 1)
				if idx < 0 or idx >= len(self.outputs):
					msg(f'Output must be in the range 1-{len(self.outputs)}')
				else:
					o_amt = self.outputs[idx].amt
					cm = ' (change output)' if chg_idx == idx else ''
					prompt = f'Fee will be deducted from output {idx+1}{cm} ({o_amt} {self.coin})'
					if check_sufficient_funds(o_amt):
						if self.cfg.yes:
							msg(prompt)
						else:
							from ..ui import keypress_confirm
							if not keypress_confirm(self.cfg, prompt+'.  OK?', default_yes=True):
								continue
						self.bump_output_idx = idx
						return idx



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>

"""
test.objattrtest_d.btc_mainnet: BTC mainnet test vectors for MMGen data objects
"""

from .common import sample_objs, seed_bin, atd
from ..include.common import cfg
from mmgen.protocol import init_proto

proto = init_proto(cfg, 'btc', need_amt=True)

from mmgen.key import PrivKey, WifKey
from mmgen.addr import CoinAddr, MMGenID, AddrIdx, MMGenAddrType, MMGenPasswordType
from mmgen.amt import BTCAmt
from mmgen.tw.shared import TwMMGenID

sample_objs.update({
	'PrivKey':   PrivKey(proto, seed_bin, compressed=True, pubkey_type='std'),
	'WifKey':    WifKey(proto, '5HwzecKMWD82ppJK3qMKpC7ohXXAwcyAN5VgdJ9PLFaAzpBG4sX'),
	'CoinAddr':  CoinAddr(proto, '1111111111111111111114oLvT2'),
	'BTCAmt':    BTCAmt('0.01'),
	'MMGenID':   MMGenID(proto, 'F00F00BB:B:1'),
	'TwMMGenID': TwMMGenID(proto, 'F00F00BB:S:23'),
})

from mmgen.addrlist import AddrListEntry
from mmgen.passwdlist import PasswordListEntry
from mmgen.obj import TwComment, CoinTxID, HexStr
from mmgen.seed import SeedID, SeedBase
from mmgen.subseed import SubSeed, SubSeedIdx
from mmgen.seedsplit import (
	SeedShareCount,
	SeedSplitIDString,
	SeedShareIdx,
	MasterShareIdx,
	SeedShareList,
	SeedShareLast,
	SeedShareMaster,
	SeedShareMasterJoining
)
from mmgen.proto.btc.tw.unspent import BitcoinTwUnspentOutputs
from mmgen.tx.base import Base

tests = {
	# addr.py
	'AddrListEntry': atd({
		'addr':          (0b01001, CoinAddr),
		'idx':           (0b01001, AddrIdx),
		'comment':       (0b01101, TwComment),
		'sec':           (0b01001, PrivKey),
#		'viewkey':       (0b01001, ViewKey),        # TODO
#		'wallet_passwd': (0b01001, WalletPassword), # TODO
		},
		(proto,),
		{}
	),
	'PasswordListEntry': atd({
		'passwd':  (0b00001, str),
		'idx':     (0b01001, AddrIdx),
		'comment': (0b01101, TwComment),
		'sec':     (0b01001, PrivKey),
		},
		(proto,),
		{'passwd': 'ΑlphaΩmega', 'idx': 1},
	),
	# obj.py
	'PrivKey': atd({
		'compressed': (0b00001, bool),
		'wif':        (0b00001, WifKey),
		},
		(proto, seed_bin),
		{'compressed': True, 'pubkey_type': 'std'},
	),
	'MMGenAddrType': atd({
		'name':        (0b01001, str),
		'pubkey_type': (0b01001, str),
		'compressed':  (0b11001, bool),
		'gen_method':  (0b11001, str),
		'addr_fmt':    (0b11001, str),
		'wif_label':   (0b11001, str),
		'extra_attrs': (0b11001, tuple),
		'desc':        (0b01001, str),
		},
		(proto, 'S'),
		{},
	),
	# seed.py
	'SeedBase': atd({
		'data': (0b00001, bytes),
		'sid':  (0b00001, SeedID),
		},
		(cfg,),
		{'seed_bin': seed_bin},
	),
	'SubSeed': atd({
		'idx':    (0b00001, int),
		'nonce':  (0b00001, int),
		'ss_idx': (0b01001, SubSeedIdx),
		},
		[sample_objs['SubSeedList'], 1, 1, 'short'],
		{},
	),
	'SeedShareList': atd({
		'count':  (0b01001, SeedShareCount),
		'id_str': (0b01001, SeedSplitIDString),
		},
		[sample_objs['Seed'], sample_objs['SeedShareCount']],
		{},
	),
	'SeedShareLast': atd({
		'idx': (0b01001, SeedShareIdx),
		},
		[sample_objs['SeedShareList']],
		{},
	),
	'SeedShareMaster': atd({
		'idx':   (0b01001, MasterShareIdx),
		'nonce': (0b00001, int),
		},
		[sample_objs['SeedShareList'], 7, 0],
		{},
	),
	'SeedShareMasterJoining': atd({
		'id_str': (0b01001, SeedSplitIDString),
		'count':  (0b01001, SeedShareCount),
		},
		[cfg, sample_objs['MasterShareIdx'], sample_objs['Seed'], 'foo', 2],
		{},
	),
	# twuo.py
	'BitcoinTwUnspentOutputs.MMGenTwUnspentOutput': atd({
		'txid':         (0b01001, CoinTxID),
		'vout':         (0b01001, int),
		'amt':          (0b01001, BTCAmt),
		'amt2':         (0b01001, BTCAmt),
		'comment':      (0b01101, TwComment),
		'twmmid':       (0b01001, TwMMGenID),
		'addr':         (0b01001, CoinAddr),
		'confs':        (0b00001, int),
		'scriptPubKey': (0b01001, HexStr),
		'skip':         (0b00101, str),
		},
		(proto,),
		{
			'amt': BTCAmt('0.01'),
			'twmmid': 'F00F00BB:B:17',
			'addr': '1111111111111111111114oLvT2',
			'confs': 100000,
			'scriptPubKey': 'ff',
		},
	),
	# tx.py
	'Base.Input': atd({
		'vout':         (0b01001, int),
		'amt':          (0b01001, BTCAmt),
		'comment':      (0b01101, TwComment),
		'mmid':         (0b01001, MMGenID),
		'addr':         (0b01001, CoinAddr),
		'confs':        (0b01001, int),
		'txid':         (0b01001, CoinTxID),
		'have_wif':     (0b00011, bool),
		'scriptPubKey': (0b01001, HexStr),
		'sequence':     (0b00001, int),
		},
		(proto,),
		{'amt': BTCAmt('0.01'), 'addr': sample_objs['CoinAddr']},
	),
	'Base.Output': atd({
		'vout':         (0b01001, int),
		'amt':          (0b01001, BTCAmt),
		'comment':      (0b01101, TwComment),
		'mmid':         (0b01001, MMGenID),
		'addr':         (0b01001, CoinAddr),
		'confs':        (0b01001, int),
		'txid':         (0b01001, CoinTxID),
		'have_wif':     (0b00011, bool),
		'is_chg':       (0b00001, bool),
		},
		(proto,),
		{'amt': BTCAmt('0.01'), 'addr': sample_objs['CoinAddr']},
	),
}

tests['MMGenPasswordType'] = atd(tests['MMGenAddrType'].attrs, [proto, 'P'], {})



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
tx.__init__: transaction class initializer
"""

def _base_proto_subclass(clsname, modname, kwargs):
	proto = kwargs['proto']
	if proto:
		clsname = ('Token' if proto.tokensym else '') + clsname
		modname = f'mmgen.proto.{proto.base_proto_coin.lower()}.tx.{modname}'
	else:
		modname = 'mmgen.tx.base'
	import importlib
	return getattr(importlib.import_module(modname), clsname)

def _get_cls_info(clsname, modname, kwargs):
	"""
	determine cls/mod/proto and pass them to _base_proto_subclass() to get a TX instance
	"""
	if 'proto' in kwargs:
		proto = kwargs['proto']
	elif 'data' in kwargs:
		proto = kwargs['data']['proto']
	elif 'filename' in kwargs:
		from .file import MMGenTxFile
		proto = MMGenTxFile.get_proto(kwargs['cfg'], kwargs['filename'], quiet_open=True)
	elif clsname == 'Base':
		proto = None
	else:
		raise ValueError(
			f"{clsname} must be instantiated with 'proto', 'data' or 'filename' keyword")

	if clsname == 'Completed':
		from ..util import get_extension, die
		from .completed import Completed
		ext = get_extension(kwargs['filename'])
		cls = Completed.ext_to_cls(ext, proto)
		if not cls:
			die(1, f'{ext!r}: unrecognized file extension for CompletedTX')
		clsname = cls.__name__
		modname = cls.__module__.rsplit('.', maxsplit=1)[-1]
	elif clsname == 'New' and kwargs['target'] == 'swaptx':
		clsname = 'NewSwap'
		modname = 'new_swap'

	kwargs['proto'] = proto

	if 'automount' in kwargs:
		if kwargs['automount']:
			clsname = 'Automount' + clsname
		del kwargs['automount']

	return (clsname, modname, kwargs)

async def _add_twctl(clsname, modname, kwargs):
	proto = kwargs['proto']
	# TwCtl instance required to retrieve the 'symbol' and 'decimals' parameters
	# of token contract (see twctl:import_token()).
	# No twctl required by the Unsigned and Signed classes used during signing,
	# or by the New and Bump classes, which already have a twctl.
	if proto and proto.tokensym and clsname in (
			'OnlineSigned',
			'AutomountOnlineSigned',
			'Sent',
			'AutomountSent'):
		from ..tw.ctl import TwCtl
		kwargs['twctl'] = await TwCtl(kwargs['cfg'], proto, no_rpc=True)
	return (clsname, modname, kwargs)

def _get(clsname, modname, kwargs):
	return _base_proto_subclass(*_get_cls_info(clsname, modname, kwargs))(**kwargs)

async def _get_async(clsname, modname, kwargs):
	return _base_proto_subclass(*(await _add_twctl(*_get_cls_info(clsname, modname, kwargs))))(**kwargs)

BaseTX         = lambda **kwargs: _get('Base',     'base',     kwargs)
NewTX          = lambda **kwargs: _get('New',      'new',      kwargs)
NewSwapTX      = lambda **kwargs: _get('NewSwap',  'new_swap', kwargs)
BumpTX         = lambda **kwargs: _get('Bump',     'bump',     kwargs)
UnsignedTX     = lambda **kwargs: _get('Unsigned', 'unsigned', kwargs)
SignedTX       = lambda **kwargs: _get('Signed',   'signed',   kwargs)

CompletedTX    = lambda **kwargs: _get_async('Completed',    'completed', kwargs)
OnlineSignedTX = lambda **kwargs: _get_async('OnlineSigned', 'online',    kwargs)
SentTX         = lambda **kwargs: _get_async('Sent',         'online',    kwargs)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
keygen: Public key generation initialization code for the MMGen suite
"""

from collections import namedtuple
from .key import PrivKey

keygen_public_data = namedtuple(
	'keygen_public_data', [
		'pubkey',
		'viewkey_bytes',
		'pubkey_type',
		'compressed'])

class keygen_base:

	def __init__(self, cfg):
		if not (self.production_safe or cfg.test_suite):
			from .util import die
			die(2,
				f'Public key generator {type(self).__name__!r} is not safe from timing attacks '
				'and may only be used in a testing environment')

	def gen_data(self, privkey):
		assert isinstance(privkey, PrivKey)
		return keygen_public_data(
			self.to_pubkey(privkey),
			self.to_viewkey(privkey),
			privkey.pubkey_type,
			privkey.compressed)

	def to_viewkey(self, privkey):
		return None

	@classmethod
	def get_clsname(cls, cfg, *, silent=False):
		return cls.__name__

backend_data = {
	'std': {
		'backends': ('libsecp256k1', 'python-ecdsa'),
		'package': 'secp256k1',
	},
	'monero': {
		'backends': ('nacl', 'ed25519ll-djbec', 'ed25519'),
		'package': 'xmr',
	},
	'zcash_z': {
		'backends': ('nacl',),
		'package': 'zec',
	},
}

def get_backends(pubkey_type):
	return backend_data[pubkey_type]['backends']

def get_pubkey_type_cls(pubkey_type):
	import importlib
	return getattr(
		importlib.import_module(f'mmgen.proto.{backend_data[pubkey_type]["package"]}.keygen'),
		'backend')

def _check_backend(cfg, backend, pubkey_type, *, desc='keygen backend'):

	from .util import is_int, die

	assert is_int(backend), f'illegal value for {desc} (must be an integer)'

	backends = get_backends(pubkey_type)

	if not (1 <= int(backend) <= len(backends)):
		die(1,
			f'{backend}: {desc} out of range\n' +
			'Configured backends: ' +
			' '.join(f'{n}:{k}' for n, k in enumerate(backends, 1))
		)

	cfg._util.qmsg(f'Using backend {backends[int(backend)-1]!r} for public key generation')

	return True

def check_backend(cfg, proto, backend, addr_type):

	from .addr import MMGenAddrType
	pubkey_type = MMGenAddrType(proto, addr_type or proto.dfl_mmtype).pubkey_type

	return  _check_backend(cfg, backend, pubkey_type, desc='--keygen-backend parameter')

def KeyGenerator(cfg, proto, pubkey_type, *, backend=None, silent=False):
	"""
	factory function returning a key generator backend for the specified pubkey type
	"""
	assert pubkey_type in proto.pubkey_types, f'{pubkey_type!r}: invalid pubkey type for coin {proto.coin}'

	pubkey_type_cls = get_pubkey_type_cls(pubkey_type)

	backend = backend or cfg.keygen_backend

	if backend:
		_check_backend(cfg, backend, pubkey_type)

	backend_id = backend_data[pubkey_type]['backends'][int(backend) - 1 if backend else 0]

	backend_clsname = getattr(
		pubkey_type_cls,
		backend_id.replace('-', '_')
			).get_clsname(cfg, silent=silent)

	return getattr(pubkey_type_cls, backend_clsname)(cfg)



#!/usr/bin/env python3

"""
test.modtest_d.swap: swap unit tests for the MMGen suite
"""

from mmgen.color import cyan
from mmgen.cfg import Config
from mmgen.amt import UniAmt
from mmgen.swap.proto.thorchain import SwapCfg, SwapAsset, Memo
from mmgen.protocol import init_proto

from ..include.common import cfg, vmsg, make_burn_addr

class unit_tests:

	def cfg(self, name, ut, desc='Swap configuration'):

		for tl_arg, tl_chk, si_arg in (
				(None,      None,             None),
				('1',       UniAmt('1'),      None),
				('33',      UniAmt('33'),     7),
				('2%',      0.98,             14),
				('-2%',     1.02,             1),
				('3.333%',  0.96667,          1),
				('-3.333%', 1.03333,          3),
				('1.2345',  UniAmt('1.2345'), 10)):
			cfg_data = {
				'trade_limit': tl_arg,
				'stream_interval': None if si_arg is None else str(si_arg)}
			sc = SwapCfg(Config(cfg_data))
			vmsg(f'  trade_limit:     {tl_arg} => {sc.trade_limit}')
			vmsg(f'  stream_interval: {si_arg} => {sc.stream_interval}')
			assert sc.trade_limit == tl_chk
			assert sc.stream_interval == sc.si.dfl if si_arg is None else si_arg
			assert sc.stream_quantity == 0

		vmsg('\n  Testing error handling')

		def bad1():
			SwapCfg(Config({'trade_limit': 'x'}))

		def bad2():
			SwapCfg(Config({'trade_limit': '1.23x'}))

		def bad3():
			SwapCfg(Config({'stream_interval': 30}))

		def bad4():
			SwapCfg(Config({'stream_interval': 0}))

		def bad5():
			SwapCfg(Config({'stream_interval': 'x'}))

		ut.process_bad_data((
			('bad1', 'SwapCfgValueError', 'invalid parameter', bad1),
			('bad2', 'SwapCfgValueError', 'invalid parameter', bad2),
			('bad3', 'SwapCfgValueError', 'invalid parameter', bad3),
			('bad4', 'SwapCfgValueError', 'invalid parameter', bad4),
			('bad5', 'SwapCfgValueError', 'invalid parameter', bad5),
		), pfx='')

		return True

	def asset(self, name, ut, desc='SwapAsset class'):
		for name, full_name, memo_name, chain, tokensym, direction in (
			('BTC',      'BTC.BTC',  'b',        'BTC', None,   'recv'),
			('LTC',      'LTC.LTC',  'l',        'LTC', None,   'recv'),
			('BCH',      'BCH.BCH',  'c',        'BCH', None,   'recv'),
			('ETH.USDT', 'ETH.USDT', 'ETH.USDT', 'ETH', 'USDT', 'recv'),
		):
			a = SwapAsset(name, direction)
			vmsg(f'  {a.name}')
			assert a.name == name
			assert a.full_name == full_name
			assert a.direction == direction
			assert a.tokensym == tokensym
			assert a.chain == chain
			assert a.memo_asset_name == memo_name
		return True

	def memo(self, name, ut, desc='Swap transaction memo'):

		for coin, addrtype, asset_name, token in (
			('ltc', 'bech32',     'LTC',      None),
			('bch', 'compressed', 'BCH',      None),
			('eth', None,         'ETH',      None),
			('eth', None,         'ETH.USDT', 'USDT'),
		):
			proto = init_proto(cfg, coin, tokensym=token, need_amt=True)
			addr = make_burn_addr(proto, addrtype)
			asset = SwapAsset(asset_name, 'recv')

			vmsg(f'\nTesting asset {cyan(asset_name)}:')

			for limit, limit_chk, si, suf in (
				('123.4567',   12340000000, None, '1234e7/3/0'),
				('1.234567',   123400000,   1,    '1234e5/1/0'),
				('0.01234567', 1234000,     10,   '1234e3/10/0'),
				('0.00012345', 12345,       20,   '12345/20/0'),
				(None,         0,           3,    '0/3/0'),
			):
				vmsg('\nTesting memo initialization:')
				swap_cfg = SwapCfg(Config({'trade_limit': limit, 'stream_interval': si}))
				m = Memo(
					swap_cfg,
					proto,
					asset,
					addr,
					trade_limit = None if limit is None else UniAmt(limit))
				vmsg(f'str(memo):  {m}')
				vmsg(f'repr(memo): {m!r}')
				vmsg(f'limit:      {limit}')

				assert str(m).endswith(':' + suf), f'{m} doesn’t end with {suf}'

				p = Memo.parse(m)
				limit_dec = UniAmt(p.trade_limit, from_unit='satoshi')
				vmsg(f'limit_dec:  {limit_dec.hl()}')

				vmsg('\nTesting memo parsing:')
				from pprint import pformat
				vmsg(pformat(p._asdict()))
				assert p.proto == 'THORChain'
				assert p.function == 'SWAP'
				assert p.asset.chain == coin.upper()
				assert p.asset.coin == token or coin.upper()
				assert p.address == addr.views[addr.view_pref]
				assert p.trade_limit == limit_chk
				assert p.stream_interval == si or swap_cfg.si.dfl, f'{p.stream_interval} != {swap_cfg.si.dfl}'
				assert p.stream_quantity == 0 # auto

			vmsg('\nTesting is_partial_memo():')
			for vec in (
				str(m),
				'SWAP:xyz',
				'=:xyz',
				's:xyz',
				'a:xz',
				'+:xz',
				'WITHDRAW:xz',
				'LOAN+:xz:x:x',
				'TRADE-:xz:x:x',
				'BOND:xz',
			):
				vmsg(f'  pass: {vec}')
				assert Memo.is_partial_memo(vec.encode('ascii')), vec

			for vec in (
				'=',
				'swap',
				'swap:',
				'swap:abc',
				'SWAP:a',
			):
				vmsg(f'  fail: {vec}')
				assert not Memo.is_partial_memo(vec.encode('ascii')), vec

			vmsg('\nTesting error handling:')

			def bad(s):
				return lambda: Memo.parse(s)

			def bad10():
				coin = 'BTC'
				proto = init_proto(cfg, coin, need_amt=True)
				addr = make_burn_addr(proto, 'C')
				asset = SwapAsset(coin, 'send')
				Memo(swap_cfg, proto, asset, addr, trade_limit=None)

			def bad11():
				SwapAsset('XYZ', 'send')

			def bad12():
				SwapAsset('DOGE', 'send')

			ut.process_bad_data((
				('bad1',  'SwapMemoParseError', 'must contain',      bad('x')),
				('bad2',  'SwapMemoParseError', 'must contain',      bad('y:z:x')),
				('bad3',  'SwapMemoParseError', 'function abbrev',   bad('z:l:foobar:0/3/0')),
				('bad4',  'SwapAssetError',     'unrecognized',      bad('=:x:foobar:0/3/0')),
				('bad5',  'SwapMemoParseError', 'failed to parse',   bad('=:l:foobar:n')),
				('bad6',  'SwapMemoParseError', 'invalid specifier', bad('=:l:foobar:x/3/0')),
				('bad7',  'SwapMemoParseError', 'extra',             bad('=:l:foobar:0/3/0:x')),
				('bad10', 'AssertionError',     'recv',              bad10),
				('bad11', 'SwapAssetError',     'unrecognized',      bad11),
				('bad12', 'SwapAssetError',     'unsupported',       bad12),
			), pfx='')

		return True
